########################################################################################################################
# Module: tests/test_mcmc_metrics.py
# Description: Tests for metrics analysing sample quality.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

import unittest

import jax.numpy as np
from jax import random
import numpy.testing as npt

from mocat.src.mcmc import metrics
from mocat.src.core import CDict
from mocat.src import kernels


class TestIsCdict(unittest.TestCase):
    arr = np.arange(20).reshape(10, 2)
    cdict = CDict(value=arr)

    def test_array(self):
        out_bool = metrics._is_cdict(self.arr)
        self.assertFalse(out_bool)

        out_bool, out_arr = metrics._is_cdict(self.arr, True)
        self.assertFalse(out_bool)
        npt.assert_array_equal(self.arr, out_arr)

    def test_cdict(self):
        out_bool = metrics._is_cdict(self.cdict)
        self.assertTrue(out_bool)

        out_bool, out_arr = metrics._is_cdict(self.cdict, True)
        self.assertTrue(out_bool)
        npt.assert_array_equal(self.arr, out_arr)


class TestAcceptanceRate(unittest.TestCase):

    all_reject_arr = np.zeros((5, 2))
    all_accept_arr = np.arange(1, 21).reshape(10, 2)
    mixed_arr = np.concatenate([all_reject_arr, all_accept_arr])

    all_accept_cdict = CDict(value=all_accept_arr)
    all_reject_cdict = CDict(value=all_reject_arr)
    mixed_cdict = CDict(value=mixed_arr)

    all_accept_cdict.alpha = np.ones(len(all_accept_arr))
    all_reject_cdict.alpha = np.zeros(len(all_reject_arr))
    mixed_cdict.alpha = np.concatenate([all_accept_cdict.alpha, all_reject_cdict.alpha])

    def test_array(self):
        accept_rate = metrics.acceptance_rate(self.all_accept_arr)
        self.assertEqual(accept_rate, 1.)

        accept_rate = metrics.acceptance_rate(self.all_reject_arr)
        self.assertEqual(accept_rate, 0.)

        accept_rate = metrics.acceptance_rate(self.mixed_arr)
        self.assertEqual(accept_rate, 10/14)

    def test_cdict(self):
        accept_rate = metrics.acceptance_rate(self.all_accept_cdict, alpha=False)
        self.assertEqual(accept_rate, 1.)

        accept_rate = metrics.acceptance_rate(self.all_reject_cdict, alpha=False)
        self.assertEqual(accept_rate, 0.)

        accept_rate = metrics.acceptance_rate(self.mixed_cdict, alpha=False)
        self.assertEqual(accept_rate, 10/14)

    def test_cdict_alpha(self):
        accept_rate = metrics.acceptance_rate(self.all_accept_cdict)
        self.assertEqual(accept_rate, 1.)

        accept_rate = metrics.acceptance_rate(self.all_reject_cdict)
        self.assertEqual(accept_rate, 0.)

        accept_rate = metrics.acceptance_rate(self.mixed_cdict)
        self.assertAlmostEqual(accept_rate, 10/15)


class TestAcceptanceRateEnsemble(unittest.TestCase):

    all_reject_arr = np.zeros((5, 3, 2))
    all_accept_arr = np.arange(1, 25).reshape(4, 3, 2)
    mixed_arr = np.concatenate([all_reject_arr, all_accept_arr])

    def test_array(self):
        accept_rate = metrics.acceptance_rate(self.all_accept_arr)
        self.assertEqual(accept_rate, 1.)

        accept_rate = metrics.acceptance_rate(self.all_reject_arr)
        self.assertEqual(accept_rate, 0.)

        accept_rate = metrics.acceptance_rate(self.mixed_arr)
        self.assertEqual(accept_rate, 4/8)


class TestAutocorrelation(unittest.TestCase):

    key = random.PRNGKey(0)
    ind_draws_arr = random.normal(key, (100,))
    ind_draws_cdict = CDict(value=ind_draws_arr)

    corr_draws_arr = 0.1 * np.cumsum(ind_draws_arr)
    corr_draws_cdict = CDict(value=corr_draws_arr)

    ind_draws_cdict_pot = CDict(value=corr_draws_arr,
                                 potential=ind_draws_arr)
    corr_draws_cdict_pot = CDict(value=ind_draws_arr,
                                 potential=corr_draws_arr)

    def test_array(self):
        ind_autocorr = metrics.autocorrelation(self.ind_draws_arr)
        self.assertEqual(ind_autocorr[0], 1.)
        npt.assert_array_equal(np.abs(ind_autocorr[1:]) < 0.3, True)

        corr_autocorr = metrics.autocorrelation(self.corr_draws_arr)
        self.assertEqual(corr_autocorr[0], 1.)
        npt.assert_array_equal(np.abs(corr_autocorr[1]) > 0.5, True)
        npt.assert_array_equal(np.abs(corr_autocorr[50]) < 0.5, True)
        npt.assert_array_equal(np.abs(corr_autocorr[-1]) < 0.3, True)

    def test_cdict(self):
        ind_autocorr = metrics.autocorrelation(self.ind_draws_cdict)
        self.assertEqual(ind_autocorr[0], 1.)
        npt.assert_array_equal(np.abs(ind_autocorr[1:]) < 0.3, True)

        corr_autocorr = metrics.autocorrelation(self.corr_draws_cdict)
        self.assertEqual(corr_autocorr[0], 1.)
        npt.assert_array_equal(np.abs(corr_autocorr[1]) > 0.5, True)
        npt.assert_array_equal(np.abs(corr_autocorr[50]) < 0.5, True)
        npt.assert_array_equal(np.abs(corr_autocorr[-1]) < 0.3, True)

    def test_cdict_pot(self):
        # Potential
        ind_autocorr = metrics.autocorrelation(self.ind_draws_cdict_pot)
        self.assertEqual(ind_autocorr[0], 1.)
        npt.assert_array_equal(np.abs(ind_autocorr[1:]) < 0.3, True)

        corr_autocorr = metrics.autocorrelation(self.corr_draws_cdict_pot)
        self.assertEqual(corr_autocorr[0], 1.)
        npt.assert_array_equal(np.abs(corr_autocorr[1]) > 0.5, True)
        npt.assert_array_equal(np.abs(corr_autocorr[50]) < 0.5, True)
        npt.assert_array_equal(np.abs(corr_autocorr[-1]) < 0.3, True)

        # Value
        ind_autocorr = metrics.autocorrelation(self.corr_draws_cdict_pot, dim=0)
        self.assertEqual(ind_autocorr[0], 1.)
        npt.assert_array_equal(np.abs(ind_autocorr[1:]) < 0.3, True)

        corr_autocorr = metrics.autocorrelation(self.ind_draws_cdict_pot, dim=0)
        self.assertEqual(corr_autocorr[0], 1.)
        npt.assert_array_equal(np.abs(corr_autocorr[1]) > 0.5, True)
        npt.assert_array_equal(np.abs(corr_autocorr[50]) < 0.5, True)
        npt.assert_array_equal(np.abs(corr_autocorr[-1]) < 0.3, True)


class testIAT(unittest.TestCase):

    key = random.PRNGKey(0)
    ind_draws_arr = random.normal(key, (100,))
    ind_draws_cdict = CDict(value=ind_draws_arr)

    corr_draws_arr = 0.1 * np.cumsum(ind_draws_arr)

    def test_array(self):
        ind_iat = metrics.integrated_autocorrelation_time(self.ind_draws_arr)
        corr_iat = metrics.integrated_autocorrelation_time(self.corr_draws_arr)

        self.assertIsInstance(ind_iat, float)
        self.assertIsInstance(corr_iat, float)
        self.assertLess(ind_iat, 2.)
        self.assertGreater(corr_iat, 8.)


class testSJD(unittest.TestCase):
    zeros_arr = np.zeros(10)
    zeros_cdict = CDict(value=zeros_arr)
    seq_arr = np.arange(10)
    seq_cdict = CDict(value=seq_arr)

    def test_array(self):
        accept_rate = metrics.squared_jumping_distance(self.zeros_arr)
        self.assertEqual(accept_rate, 0.)

        accept_rate = metrics.squared_jumping_distance(self.seq_arr)
        self.assertEqual(accept_rate, 1.)

    def test_cdict(self):
        accept_rate = metrics.squared_jumping_distance(self.zeros_cdict)
        self.assertEqual(accept_rate, 0.)

        accept_rate = metrics.squared_jumping_distance(self.seq_cdict)
        self.assertEqual(accept_rate, 1.)


class testKSDStdGaussian(unittest.TestCase):

    key = random.PRNGKey(0)
    dim = 2

    n_small = 10
    ind_draws_arr_n_small = random.normal(key, (n_small, dim))
    sample_n_small = CDict(value=ind_draws_arr_n_small, grad_potential=ind_draws_arr_n_small)

    n_large = 1000
    ind_draws_arr_n_large = random.normal(key, (n_large, dim))
    sample_n_large = CDict(value=ind_draws_arr_n_large, grad_potential=ind_draws_arr_n_large)

    def testksd_gaussian_kernel(self):
        kernel = kernels.Gaussian(bandwidth=1.)
        ksd_n_small_a = metrics.ksd(self.sample_n_small, kernel)
        ksd_n_large_a = metrics.ksd(self.sample_n_large, kernel)

        self.assertIsInstance(ksd_n_small_a, float)
        self.assertIsInstance(ksd_n_large_a, float)
        self.assertLess(ksd_n_large_a, ksd_n_small_a)

        kernel.parameters.bandwidth = 10.
        ksd_n_small_b = metrics.ksd(self.sample_n_small, kernel)
        ksd_n_large_b = metrics.ksd(self.sample_n_large, kernel)

        self.assertIsInstance(ksd_n_small_b, float)
        self.assertIsInstance(ksd_n_large_b, float)
        self.assertLess(ksd_n_large_b, ksd_n_small_b)

    def testksd_IMQ_kernel(self):
        kernel = kernels.IMQ(bandwidth=1., c=1., beta=-0.5)
        ksd_n_small_a = metrics.ksd(self.sample_n_small, kernel)
        ksd_n_large_a = metrics.ksd(self.sample_n_large, kernel)

        self.assertIsInstance(ksd_n_small_a, float)
        self.assertIsInstance(ksd_n_large_a, float)
        self.assertLess(ksd_n_large_a, ksd_n_small_a)

        kernel.parameters.bandwidth = 10.
        kernel.parameters.c = 0.1
        kernel.parameters.beta = -0.1

        ksd_n_small_b = metrics.ksd(self.sample_n_small, kernel)
        ksd_n_large_b = metrics.ksd(self.sample_n_large, kernel)

        self.assertIsInstance(ksd_n_small_b, float)
        self.assertIsInstance(ksd_n_large_b, float)
        self.assertLess(ksd_n_large_b, ksd_n_small_b)


if __name__ == '__main__':
    unittest.main()



########################################################################################################################
# Module: mcmc/run.py
# Description: Run Markov chains!
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from time import time
from typing import Union, Tuple, Type, Any
from inspect import isclass

from jax import numpy as np, jit
from jax.lax import scan

from mocat.src.core import Scenario, CDict
from mocat.src.mcmc.corrections import Correction, Uncorrected
from mocat.src.mcmc.sampler import MCMCSampler


def startup_mcmc(scenario: Scenario,
                 sampler: MCMCSampler,
                 random_key: Union[None, np.ndarray],
                 correction: Union[None, str, Correction, Type[Correction]],
                 **kwargs) -> Tuple[MCMCSampler, Correction]:
    # Setup correction
    if correction == 'sampler_default':
        correction = sampler.default_correction

    if correction is None:
        correction = Uncorrected()
    elif isclass(correction) and issubclass(correction, Correction):
        correction = correction(**kwargs)
    elif not isinstance(correction, Correction):
        raise TypeError(f'Correction must be of type mocat.Correction')

    # Update kwargs
    for key, value in kwargs:
        if hasattr(sampler, key):
            setattr(sampler, key, value)
        if hasattr(correction, key):
            setattr(correction, key, value)

    # Startup
    sampler.startup(scenario, random_key)
    correction.startup(scenario, sampler)

    if None in sampler.parameters.__dict__.values():
        raise ValueError(f'None found in {sampler.name}.parameters: \n{sampler.parameters}')

    # random_key = None -> use last key from previous run, otherwise set given random_key
    if random_key is not None:
        sampler.initial_extra.random_key = random_key
    sampler.initial_extra.iter = 0

    return sampler, correction


def mcmc_run_params(sampler, correction):
    return CDict(name=sampler.name,
                 parameters=sampler.parameters.copy(),
                 correction=correction.__class__.__name__,
                 tuning=sampler.tuning,
                 initial_state=sampler.initial_state.copy(),
                 initial_extra=sampler.initial_extra.copy())


def run_mcmc(scenario: Scenario,
             sampler: MCMCSampler,
             n: int,
             random_key: Union[None, np.ndarray],
             correction: Union[None, str, Correction, Type[Correction]] = 'sampler_default',
             name: str = None,
             return_random_key: bool = False,
             **kwargs) -> Union[CDict, Tuple[CDict, np.ndarray]]:
    sampler, correction = startup_mcmc(scenario, sampler, random_key, correction, **kwargs)

    run_params = mcmc_run_params(sampler, correction)

    @jit
    def markov_kernel(previous_carry: Tuple[CDict, CDict],
                      _: Any) -> Tuple[Tuple[CDict, CDict], CDict]:
        previous_state, previous_extra = previous_carry
        reject_state = previous_state.copy()
        reject_extra = previous_extra.copy()
        reject_extra.iter += 1

        reject_state, reject_extra = sampler.always(scenario,
                                                    reject_state,
                                                    reject_extra)

        proposed_state, proposed_extra = sampler.proposal(scenario,
                                                          reject_state,
                                                          reject_extra)

        corrected_state, corrected_extra = correction(scenario, sampler,
                                                      reject_state, reject_extra,
                                                      proposed_state, proposed_extra)

        return (corrected_state, corrected_extra), corrected_state

    start = time()

    final_carry, chain = scan(markov_kernel,
                              (sampler.initial_state, sampler.initial_extra),
                              None,
                              length=n)

    end = time()

    sampler.parameters = final_carry[1].parameters

    chain.run_params = run_params

    if sampler.update_initial_state:
        sampler.initial_state = final_carry[0]

    if sampler.update_initial_extra:
        sampler.initial_extra = final_carry[1]

    chain.time = end - start

    if name is None:
        name = scenario.name + ": " + sampler.name
    chain.name = name

    return (chain, final_carry[1].random_key) if return_random_key else chain



########################################################################################################################
# Module: mcmc/standard_mcmc.py
# Description: Some common MCMC proposals.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Tuple, Union
from functools import partial

import jax.numpy as np
from jax import random

from mocat.src.core import CDict, Scenario
from mocat.src.mcmc.sampler import MCMCSampler, mh_acceptance_probability
from mocat.src import utils
from mocat.src.mcmc.corrections import Metropolis


# Random-walk
# Identity pre-conditioner
class RandomWalk(MCMCSampler):
    name = 'Random Walk'
    default_correction = Metropolis()

    def __init__(self,
                 stepsize: float = None):
        super().__init__()
        self.parameters.stepsize = stepsize
        self.tuning.target = 0.234

    def proposal(self,
                        scenario: Scenario,
                        reject_state: CDict,
                        reject_extra: CDict) -> Tuple[CDict, CDict]:
        proposed_state = reject_state.copy()

        d = scenario.dim
        x = reject_state.value

        stepsize = reject_extra.parameters.stepsize

        reject_extra.random_key, subkey = random.split(reject_extra.random_key)

        proposed_state.value = x + np.sqrt(stepsize) * random.normal(subkey, (d,))

        return proposed_state, reject_extra

    def acceptance_probability(self,
                               scenario: Scenario,
                               reject_state: CDict, reject_extra: CDict,
                               proposed_state: CDict, proposed_extra: CDict) -> Union[float, np.ndarray]:
        return np.minimum(1., np.exp(- proposed_state.potential
                                     + reject_state.potential))


# Euler-Maruyama discretisation of Overdamped Langevin dynamics
# Identity pre-conditioner
class Overdamped(MCMCSampler):
    name = 'Overdamped'
    default_correction = Metropolis()

    def __init__(self,
                 stepsize: float = None):
        super().__init__()
        self.parameters.stepsize = stepsize
        self.tuning.target = 0.574

    def startup(self,
                scenario: Scenario,
                random_key: np.ndarray):
        super().startup(scenario, random_key)
        self.initial_state.grad_potential = scenario.grad_potential(self.initial_state.value)

    def proposal(self,
                        scenario: Scenario,
                        reject_state: CDict,
                        reject_extra: CDict) -> Tuple[CDict, CDict]:
        proposed_state = reject_state.copy()

        d = scenario.dim
        x = reject_state.value
        x_grad_potential = reject_state.grad_potential

        stepsize = reject_extra.parameters.stepsize

        reject_extra.random_key, subkey = random.split(reject_extra.random_key)

        proposed_state.value = x \
                               - stepsize * x_grad_potential \
                               + np.sqrt(2 * stepsize) * random.normal(subkey, (d,))
        proposed_state.grad_potential = scenario.grad_potential(proposed_state.value)

        return proposed_state, reject_extra

    def proposal_potential(self,
                           scenario: Scenario,
                           reject_state: CDict, reject_extra: CDict,
                           proposed_state: CDict, proposed_extra: CDict) -> Union[float, np.ndarray]:
        d = scenario.dim
        stepsize = reject_extra.parameters.stepsize

        return utils.gaussian_potential(proposed_state.value,
                                        reject_state.value - stepsize * reject_state.grad_potential,
                                        np.eye(d) / (2 * stepsize))

    acceptance_probability = mh_acceptance_probability


# Hamiltonian Monte Carlo
# Identity pre-conditioner/mass matrix
class HMC(MCMCSampler):
    name = 'HMC'
    default_correction = Metropolis()

    def __init__(self,
                 stepsize: float = None,
                 leapfrog_steps: int = None):
        super().__init__()
        self.parameters.stepsize = stepsize
        self.parameters.leapfrog_steps = leapfrog_steps
        self.tuning.target = 0.651

    def startup(self,
                scenario: Scenario,
                random_key: np.ndarray):
        super().startup(scenario, random_key)
        self.initial_state.grad_potential = scenario.grad_potential(self.initial_state.value)
        if not hasattr(self.initial_state, 'momenta') or self.initial_state.momenta.shape[-1] != scenario.dim:
            self.initial_state.momenta = np.zeros(scenario.dim)

    def proposal(self,
                        scenario: Scenario,
                        reject_state: CDict,
                        reject_extra: CDict) -> Tuple[CDict, CDict]:
        d = scenario.dim

        stepsize = reject_extra.parameters.stepsize

        reject_extra.random_key, subkey = random.split(reject_extra.random_key)

        reject_state.momenta = random.normal(subkey, (d,))

        proposed_state = utils.leapfrog(reject_state,
                                        scenario.grad_potential,
                                        stepsize,
                                        self.parameters.leapfrog_steps)[-1]

        # Techinically we should reverse momenta now
        # but momenta target is symmetric and then immediately resampled at the next step anyway

        return proposed_state, reject_extra

    def acceptance_probability(self,
                               scenario: Scenario,
                               reject_state: CDict, reject_extra: CDict,
                               proposed_state: CDict, proposed_extra: CDict) -> Union[float, np.ndarray]:
        pre_min_alpha = np.exp(- proposed_state.potential
                               + reject_state.potential
                               - utils.gaussian_potential(proposed_state.momenta)
                               + utils.gaussian_potential(reject_state.momenta))

        return np.minimum(1., pre_min_alpha)


# Underdamped Langevin
# Identity pre-conditioner/mass matrix
class Underdamped(MCMCSampler):
    name = 'Underdamped'
    default_correction = Metropolis()

    def __init__(self,
                 stepsize: float = None,
                 leapfrog_steps: int = None,
                 friction: float = 1.0):
        super().__init__()
        self.parameters.stepsize = stepsize
        self.parameters.leapfrog_steps = leapfrog_steps
        self.parameters.friction = friction
        self.tuning.target = 0.651

    def startup(self,
                scenario: Scenario,
                random_key: np.ndarray):
        super().startup(scenario, random_key)
        self.initial_state.grad_potential = scenario.grad_potential(self.initial_state.value)
        if not hasattr(self.initial_state, 'momenta') or self.initial_state.momenta.shape[-1] != scenario.dim:
            self.initial_state.momenta = np.zeros(scenario.dim)

    def always(self,
               scenario: Scenario,
               reject_state: CDict,
               reject_extra: CDict) -> Tuple[CDict, CDict]:
        d = scenario.dim

        stepsize = reject_extra.parameters.stepsize
        friction = reject_extra.parameters.friction

        reject_state.momenta *= -1

        # Update p - exactly according to solution of OU process
        # Accepted even if leapfrog step is rejected
        reject_extra.random_key, subkey = random.split(reject_extra.random_key)
        reject_state.momenta = reject_state.momenta * np.exp(- friction * stepsize) \
                               + np.sqrt(1 - np.exp(- 2 * friction * stepsize)) * random.normal(subkey, (d,))
        return reject_state, reject_extra

    def proposal(self,
                        scenario: Scenario,
                        reject_state: CDict,
                        reject_extra: CDict) -> Tuple[CDict, CDict]:
        
        stepsize = reject_extra.parameters.stepsize
        
        proposed_state = utils.leapfrog(reject_state,
                                        scenario.grad_potential,
                                        stepsize,
                                        self.parameters.leapfrog_steps)[-1]
        proposed_state.momenta *= -1

        return proposed_state, reject_extra

    acceptance_probability = HMC.acceptance_probability


# Tamed Euler-Maruyama discretisation of Overdamped Langevin dynamics
# Identity pre-conditioner
class TamedOverdamped(MCMCSampler):
    name = 'Tamed Overdamped'
    default_correction = Metropolis()

    def __init__(self,
                 stepsize: float = None):
        super().__init__()
        self.parameters.stepsize = stepsize
        self.tuning.target = 0.574
        self.taming = self.taming_coord

    def startup(self,
                scenario: Scenario,
                random_key: np.ndarray):
        super().startup(scenario, random_key)
        self.initial_state.grad_potential = scenario.grad_potential(self.initial_state.value)

    def proposal(self,
                        scenario: Scenario,
                        reject_state: CDict,
                        reject_extra: CDict) -> Tuple[CDict, CDict]:
        proposed_state = reject_state.copy()

        d = scenario.dim
        x = reject_state.value
        x_grad_potential = reject_state.grad_potential

        stepsize = reject_extra.parameters.stepsize

        reject_extra.random_key, subkey = random.split(reject_extra.random_key)

        proposed_state.value = x \
                               - stepsize * x_grad_potential * self.taming(reject_state, reject_extra) \
                               + np.sqrt(2 * stepsize) * random.normal(subkey, (d,))
        proposed_state.grad_potential = scenario.grad_potential(proposed_state.value)

        return proposed_state, reject_extra

    @staticmethod
    def taming_global(state: CDict,
                      extra: CDict) -> Union[float, np.ndarray]:
        return 1. / (1 + extra.parameters.stepsize * np.linalg.norm(state.grad_potential))

    @staticmethod
    def taming_coord(state: CDict,
                     extra: CDict) -> Union[float, np.ndarray]:
        return 1. / (1 + extra.parameters.stepsize * np.abs(state.grad_potential))

    def proposal_potential(self,
                           scenario: Scenario,
                           reject_state: CDict, reject_extra: CDict,
                           proposed_state: CDict, proposed_extra: CDict) -> Union[float, np.ndarray]:
        d = scenario.dim
        stepsize = reject_extra.parameters.stepsize

        return utils.gaussian_potential(proposed_state.value,
                                        reject_state.value - stepsize * reject_state.grad_potential
                                        * self.taming(reject_state, reject_extra),
                                        np.eye(d) / (2 * stepsize))

    acceptance_probability = mh_acceptance_probability




"""Mocat: Monte Carlo Testbed"""


from mocat import utils
from mocat import mcmc
from mocat import kernels
from mocat import scenarios
from mocat import twodim

from mocat.src.core import Scenario
from mocat.src.core import CDict
from mocat.src.core import Sampler

from mocat.src.mcmc.metrics import acceptance_rate
from mocat.src.mcmc.metrics import autocorrelation
from mocat.src.mcmc.metrics import integrated_autocorrelation_time
from mocat.src.mcmc.metrics import ess
from mocat.src.mcmc.metrics import ess_per_second
from mocat.src.mcmc.metrics import squared_jumping_distance
from mocat.src.mcmc.metrics import ksd
from mocat.src.mcmc.metrics import autocorrelation_plot
from mocat.src.mcmc.metrics import trace_plot
from mocat.src.mcmc.metrics import plot_2d_samples
from mocat.src.mcmc.metrics import hist_1d_samples

from mocat.src.mcmc.sampler import MCMCSampler
from mocat.src.mcmc.run import run_mcmc

from mocat.src.mcmc.corrections import Correction
from mocat.src.mcmc.corrections import Uncorrected
from mocat.src.mcmc.corrections import Metropolis
from mocat.src.mcmc.corrections import RMMetropolis

from mocat.src.mcmc.standard_mcmc import RandomWalk
from mocat.src.mcmc.standard_mcmc import Overdamped
from mocat.src.mcmc.standard_mcmc import HMC
from mocat.src.mcmc.standard_mcmc import Underdamped
from mocat.src.mcmc.standard_mcmc import TamedOverdamped

from mocat.src.mcmc.ensemble_mcmc import EnsembleRWMH
from mocat.src.mcmc.ensemble_mcmc import EnsembleOverdamped

try:
  del src
except NameError:
  pass




########################################################################################################################
# Module: metrics.py
# Description: Some common metrics to assess sample quality.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Union, Tuple

import jax.numpy as np
from jax import vmap
import numpy as onp         # For onp.fft.fft(x, n=n) and onp.unique(x, axis=0) (JAX doesn't support as of writing)
import matplotlib.pyplot as plt

from mocat.src.core import CDict
from mocat.src.utils import _metric_plot_decorate
from mocat.src.kernels import Kernel


def _is_cdict(input: Union[np.ndarray, CDict],
              value_out: bool = False) -> Union[bool, Tuple[bool, np.ndarray]]:
    if isinstance(input, CDict):
        return (True, input.value) if value_out else True
    elif isinstance(input, np.ndarray):
        return (False, input) if value_out else False
    else:
        raise ValueError("Input isn't ndarray or CDict")


def _acceptance_rate_values(array: np.ndarray) -> float:
    if array.ndim == 1:
        return (len(np.unique(array)) - 1) / (len(array) - 1)
    elif array.ndim == 2:
        return _acceptance_rate_values(array[:, 0])
    elif array.ndim == 3:
        return (len(onp.unique(array[:, :, 0], axis=0)) - 1) / (array.shape[0] - 1)
    else:
        raise TypeError("Can't compute acceptance rate of object")


def acceptance_rate(sample: Union[np.ndarray, CDict],
                    alpha: bool = True) -> float:

    _, vals = _is_cdict(sample, True)
    if hasattr(sample, 'alpha') and alpha:
        return float(sample.alpha.mean())
    else:
        return _acceptance_rate_values(vals)


def _next_pow_two(n: int) -> int:
    i = 1
    while i < n:
        i = i << 1
    return i


def _autocorrelation_1darray(x: Union[np.ndarray, CDict],
                             max_lag_eval: int) -> np.ndarray:
    if x.ndim != 1:
        raise ValueError("_autocorrelation_1darray must be applied to a 1 dimensional np.ndarray")
    max_lag_eval = min(max_lag_eval, len(x))
    n = _next_pow_two(max_lag_eval)
    f = onp.fft.fft(x - np.mean(x), n=2 * n)
    acf = np.fft.ifft(f * np.conjugate(f))[:max_lag_eval].real
    return acf / acf[0]


def autocorrelation(sample: Union[CDict, np.ndarray],
                    dim: int = None,
                    max_lag_eval: int = 1000,
                    ensemble_index: int = None) -> np.ndarray:

    _, vals = _is_cdict(sample, value_out=True)

    if vals.ndim == 3:
        if isinstance(sample, CDict) and hasattr(sample, 'potential'):
            if ensemble_index is None:
                if dim is None:
                    return _autocorrelation_1darray(np.mean(sample.potential, axis=1), max_lag_eval)
                else:
                    return _autocorrelation_1darray(np.mean(vals[:, :, dim], axis=1), max_lag_eval)
            else:
                return autocorrelation(vals[:, ensemble_index, :], dim, max_lag_eval)

        elif isinstance(vals, np.ndarray):
            if ensemble_index is None:
                if dim is None:
                    return _autocorrelation_1darray(np.mean(sample[:, :, 0], axis=1), max_lag_eval)
                else:
                    return _autocorrelation_1darray(np.mean(sample[:, :, dim], axis=1), max_lag_eval)
            else:
                return autocorrelation(sample[:, ensemble_index, :], dim, max_lag_eval)

    elif vals.ndim == 2:
        if isinstance(sample, CDict) and hasattr(sample, 'potential'):
            if dim is None:
                return _autocorrelation_1darray(sample.potential, max_lag_eval)
            else:
                return _autocorrelation_1darray(vals[:, dim], max_lag_eval)

        elif isinstance(vals, np.ndarray):
            if dim is None:
                return _autocorrelation_1darray(vals[:, 0], max_lag_eval)
            else:
                return _autocorrelation_1darray(vals[:, dim], max_lag_eval)

    elif vals.ndim == 1:
        if isinstance(sample, CDict) and hasattr(sample, 'potential'):
            if dim is None:
                return _autocorrelation_1darray(sample.potential, max_lag_eval)
            else:
                return _autocorrelation_1darray(vals, max_lag_eval)

        elif isinstance(vals, np.ndarray):
            return _autocorrelation_1darray(vals, max_lag_eval)

    else:
        raise ValueError("Input to autocorrelation must be numpy.ndarray or Sample")


def integrated_autocorrelation_time(sample: Union[np.ndarray, CDict],
                                    dim: int = None,
                                    max_lag_iat: int = None,
                                    max_lag_eval: int = 1000,
                                    ensemble_index: int = None) -> float:

    autocorr = autocorrelation(sample, dim, max_lag_eval, ensemble_index)
    iats_all = 1 + 2 * np.cumsum(autocorr[1:])

    # https://dfm.io/posts/autocorr/
    if max_lag_iat is None:
        c = 5
        m_minus_c_tau = np.arange(1, len(iats_all) + 1) - c * iats_all
        max_lag_iat = np.argmax(m_minus_c_tau > 0)

    return float(iats_all[max_lag_iat])


def ess(sample: Union[np.ndarray, CDict],
        dim: int = None,
        max_lag_iat: int = None,
        max_lag_eval: int = 1000,
        ensemble_index: int = None) -> float:

    _, vals = _is_cdict(sample, value_out=True)

    # https://dfm.io/posts/autocorr/
    iat = integrated_autocorrelation_time(sample, dim, max_lag_iat, max_lag_eval, ensemble_index)

    n_mult = np.prod(vals.shape[:-1]) if vals.ndim == 3 and ensemble_index is None else len(vals)

    return n_mult / iat


def ess_per_second(sample: CDict,
                   dim: int = None,
                   max_lag_iat: int = None,
                   max_lag_eval: int = 1000) -> float:

    if not hasattr(sample, 'time'):
        raise ValueError("Time not stored in CDict")
    else:
        return ess(sample, dim, max_lag_iat, max_lag_eval) / sample.time


def squared_jumping_distance(sample: Union[np.ndarray, CDict],
                             dim: int = 0) -> float:

    _, vals = _is_cdict(sample, value_out=True)
    sjd = vals[1:] - vals[:-1]
    sjd = sjd ** 2
    return float(sjd[..., dim].mean())


def ksd(sample: CDict,
        kernel: Kernel,
        weighted: bool = True) -> float:

    xs = sample.value
    grad_potentials = sample.grad_potential

    n = len(xs)

    if hasattr(sample, 'weights') is None and weighted:
        weights = sample.weights
    elif hasattr(sample, 'logit_weights') and weighted:
        weights = np.exp(sample.logit_weights)
        weights /= weights.sum()
    else:
        weights = np.ones(n) / n

    def k_0_inds(x_i, y_i):
        kern_diag_grad_xy_mat = np.sum(kernel.diag_grad_xy(xs[x_i], xs[y_i]))
        grad_kx_grad_py_mat = np.dot(kernel.grad_x(xs[x_i], xs[y_i]), grad_potentials[y_i])
        grad_ky_grad_px_mat = np.dot(grad_potentials[x_i], kernel.grad_y(xs[x_i], xs[y_i]))
        kern_grad_pxy = kernel(xs[x_i], xs[y_i]) * np.dot(grad_potentials[x_i], grad_potentials[y_i])

        return (kern_diag_grad_xy_mat + grad_kx_grad_py_mat + grad_ky_grad_px_mat + kern_grad_pxy) \
               * weights[x_i] * weights[y_i]

    return float(np.sqrt((vmap(k_0_inds)(np.arange(n), np.arange(n))).sum()))


@_metric_plot_decorate
def autocorrelation_plot(sample: Union[np.ndarray, CDict],
                         dim: int = None,
                         max_lag_plot: int = 100,
                         max_lag_eval: int = 1000,
                         ax: plt.Axes = None,
                         **kwargs):
    autocor = autocorrelation(sample, dim, max_lag_eval)
    ax.bar(range(1, 1 + len(autocor)), autocor, **kwargs)
    ax.set_xlim(0, max_lag_plot)


@_metric_plot_decorate
def trace_plot(sample: Union[np.ndarray, CDict],
               dim: int = None,
               last_n: int = None,
               ax: plt.Axes = None,
               **kwargs):
    _, vals = _is_cdict(sample, True)

    if last_n is None:
        start = 0
    else:
        start = vals.shape[0] - last_n

    if dim is None and isinstance(sample, CDict) and hasattr(sample, 'potential'):
        ax.plot(sample.potential[start:], **kwargs)
    else:
        if dim is None:
            dim = 0
        ax.plot(vals[start:, ..., dim], **kwargs)


@_metric_plot_decorate
def plot_2d_samples(sample: Union[np.ndarray, CDict],
                    dim1: int = 0,
                    dim2: int = 1,
                    s: float = 0.5,
                    ax: plt.Axes = None,
                    **kwargs):
    _, vals = _is_cdict(sample, True)

    if vals.shape[-1] == 1:
        raise TypeError('Samples are one dimensional - try mocat.hist_1d_samples')
    else:
        if vals.ndim == 3:
            vals = np.concatenate(vals)

        ax.scatter(vals[:, dim1], vals[:, dim2], s, **kwargs)


@_metric_plot_decorate
def hist_1d_samples(sample: Union[np.ndarray, CDict],
                    dim: int = 0,
                    bins: int = 50,
                    density: bool = True,
                    ax: plt.Axes = None,
                    **kwargs):
    _, vals = _is_cdict(sample, True)

    if vals.ndim == 3:
        vals = np.concatenate(vals)

    ax.hist(vals[:, dim], bins=bins, density=density, **kwargs)





from mocat.src.scenarios.twodim.vectorise import TwoDimScenario

from mocat.src.scenarios.twodim.toy_scenarios import Gaussian
from mocat.src.scenarios.twodim.toy_scenarios import Banana
from mocat.src.scenarios.twodim.toy_scenarios import NealFunnel
from mocat.src.scenarios.twodim.toy_scenarios import GaussianMixture
from mocat.src.scenarios.twodim.toy_scenarios import DoubleWell
from mocat.src.scenarios.twodim.toy_scenarios import Rastrigin

from mocat.src.scenarios.twodim.visualise import visualise
from mocat.src.scenarios.twodim.visualise import RunVis
from mocat.src.scenarios.twodim.visualise import RunVisUtils






########################################################################################################################
# Module: core.py
# Description: mocat primitives a Scenario, CDict and Sampler.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from __future__ import annotations
from functools import partial
from typing import Union
from pathlib import Path
import pickle
import copy

import jax.numpy as np
from jax import grad, jit
from jax.tree_util import register_pytree_node_class
from jax.util import unzip2


class Scenario:
    name = None
    dim = None

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            if key in self.__dict__.keys():
                self.__dict__[key] = value

        if self.dim == 1:
            self._potential = self.potential

            def potential(x):
                x = np.atleast_1d(x)
                x = np.reshape(x, (x.size, 1))
                return np.squeeze(self._potential(x))

            self.potential = potential

        self.potential = jit(self.potential, static_argnums=(0,))
        self.dens = jit(self.dens, static_argnums=(0,))
        self.grad_potential = jit(grad(self.potential), static_argnums=(0,))

    def __repr__(self):
        return f"mocat.Scenario.{self.__class__.__name__}({self.__dict__.__repr__()})"

    def copy(self) -> Scenario:
        return copy.deepcopy(self)

    def potential(self, x: np.ndarray) -> Union[float, np.ndarray]:
        raise NotImplementedError(f'{self.name} potential not initiated')

    def dens(self, x: np.ndarray) -> Union[float, np.ndarray]:
        return np.exp(-self.potential(x))


@register_pytree_node_class
class CDict:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def copy(self) -> CDict:
        return CDict(**self.__dict__)

    def __repr__(self):
        return f"mocat.CDict({self.__dict__.__repr__()})"

    def save(self,
             path: Union[str, Path],
             overwrite: bool = False):
        save_CDict(self, path, overwrite)

    def tree_flatten(self):
        return unzip2(self.__dict__.items())[::-1]

    @classmethod
    def tree_unflatten(cls, keys, xs):
        return cls(**dict(zip(keys, xs)))

    def __getitem__(self,
                    item: Union[int, slice, np.ndarray]) -> CDict:
        if isinstance(item, int):
            return self.__get_int(item)
        else:
            return self.__get_general(item)

    @partial(jit, static_argnums=(0,))
    def __get_int(self,
                  int_item: int) -> CDict:
        return self.__get_general(int_item)

    def __get_general(self,
                      item: Union[str, int, slice, np.ndarray]) -> CDict:
        if isinstance(item, str):
            return self.__dict__[item]

        out_cdict = self.copy()
        for key, attr in out_cdict.__dict__.items():
            if isinstance(attr, np.ndarray):
                out_cdict.__setattr__(key, self.__dict__[key][item])
        return out_cdict

    def __add__(self,
                other: CDict) -> CDict:
        out_cdict = self.copy()
        for key, attr in out_cdict.__dict__.items():
            if isinstance(attr, np.ndarray) and hasattr(other, key):
                attr_atl = attr
                other_attr_atl = other.__dict__[key]
                out_cdict.__setattr__(key, np.append(attr_atl,
                                                     other_attr_atl, axis=0))
        if hasattr(self, 'time') and hasattr(other, 'time'):
            out_cdict.time = self.time + other.time
        return out_cdict

    def keys(self):
        return self.__dict__.keys()

    def __iter__(self):
        return self.__dict__.__iter__()


def save_CDict(CDict: CDict,
               path: Union[str, Path],
               overwrite: bool = False):
    path = Path(path)
    if path.suffix != '.CDict':
        path = path.with_suffix('.CDict')
    path.parent.mkdir(parents=True, exist_ok=True)
    if path.exists():
        if overwrite:
            path.unlink()
        else:
            raise RuntimeError(f'File {path} already exists.')
    with open(path, 'wb') as file:
        pickle.dump(CDict, file)


def load_CDict(path: Union[str, Path]) -> CDict:
    path = Path(path)
    if not path.is_file():
        raise ValueError(f'Not a file: {path}')
    if path.suffix != '.CDict':
        raise ValueError(f'Not a .CDict file: {path}')
    with open(path, 'rb') as file:
        data = pickle.load(file)

    # Convert to DeviceArray (default loads as np.ndarray, although I expect JAX to update this at some point)
    for attr_key, attr in data.__dict__.items():
        if isinstance(attr, np.ndarray):
            data.__dict__[attr_key] = np.asarray(attr)

    return data


class Sampler:

    def __init__(self,
                 name: str = None,
                 **kwargs):

        if not hasattr(self, 'parameters'):
            self.parameters = CDict()

        if name is not None:
            self.name = name
        elif not hasattr(self, 'name'):
            self.name = "Sampler"

        if self.parameters is not None:
            self.parameters.__dict__.update(kwargs)

    def __repr__(self):
        return f"mocat.Sampler.{self.__class__.__name__}({self.__dict__.__repr__()})"

    def copy(self) -> Sampler:
        return copy.deepcopy(self)

    def startup(self,
                scenario: Scenario,
                random_key: np.ndarray):
        pass





from mocat.src.mcmc.sampler import MCMCSampler
from mocat.src.mcmc.sampler import default_initial_state
from mocat.src.mcmc.sampler import default_initial_extra
from mocat.src.mcmc.sampler import mh_acceptance_probability
from mocat.src.mcmc.run import run_mcmc

from mocat.src.mcmc.metrics import acceptance_rate
from mocat.src.mcmc.metrics import autocorrelation
from mocat.src.mcmc.metrics import integrated_autocorrelation_time
from mocat.src.mcmc.metrics import ess
from mocat.src.mcmc.metrics import ess_per_second
from mocat.src.mcmc.metrics import squared_jumping_distance
from mocat.src.mcmc.metrics import autocorrelation_plot
from mocat.src.mcmc.metrics import trace_plot
from mocat.src.mcmc.metrics import plot_2d_samples
from mocat.src.mcmc.metrics import hist_1d_samples

from mocat.src.mcmc.corrections import Correction
from mocat.src.mcmc.corrections import Uncorrected
from mocat.src.mcmc.corrections import Metropolis
from mocat.src.mcmc.corrections import RMMetropolis

from mocat.src.mcmc.standard_mcmc import RandomWalk
from mocat.src.mcmc.standard_mcmc import Overdamped
from mocat.src.mcmc.standard_mcmc import HMC
from mocat.src.mcmc.standard_mcmc import Underdamped
from mocat.src.mcmc.standard_mcmc import TamedOverdamped

from mocat.src.mcmc.ensemble_mcmc import EnsembleRWMH
from mocat.src.mcmc.ensemble_mcmc import EnsembleOverdamped

from mocat.src.tests.test_mcmc_sampling import TestMetropolisCorrelatedGaussian





########################################################################################################################
# Module: visualise.py
# Description: Visualisations of MCMC on two dimensional scenarios.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Union, Tuple, Type, Any
from inspect import isclass
from time import time

import jax.numpy as np
from jax import jit
from jax.lax import scan
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.contour import QuadContourSet

from mocat.src.core import CDict, Sampler
from mocat.src.scenarios.twodim.vectorise import _generate_plot_grid, TwoDimScenario
from mocat.src.mcmc.sampler import MCMCSampler
from mocat.src.mcmc.corrections import Correction, Uncorrected, Metropolis
from mocat.src.mcmc.run import startup_mcmc, mcmc_run_params


class RunVisUtils:
    def __init__(self):
        self.plot_scen_potential = False
        self.background_samples_size = 4
        self.background_samples_alpha = 0.9
        self.proposal_size = 20
        self.max_weighted_size = 300
        self.scenario_alpha = 0.4
        self.gradient_alpha = 0.3

    def scenario_contours(self, ax, scenario, potential=False):
        return scenario.plot(ax=ax, alpha=self.scenario_alpha, zorder=0, potential=potential)

    def full_state_points(self, ax, sample):
        sample = np.atleast_2d(sample)
        if sample.ndim == 3:
            sample = np.concatenate(sample)
        return ax.scatter(sample[:, 0], sample[:, 1],
                          s=self.background_samples_size, alpha=self.background_samples_alpha, color='grey', zorder=1)

    def live_state_points(self, ax, state, weights=None):
        two_dim_state = np.atleast_2d(state)
        if weights is None:
            s = self.proposal_size
        else:
            s = weights / np.sum(weights) * self.max_weighted_size
        return ax.scatter(two_dim_state[:, 0], two_dim_state[:, 1], s=s, color='blue', zorder=4)

    @staticmethod
    def proposal_contours(ax, x, y, z):
        z = np.where(z < 1e-10, 1e-10, z)
        return ax.contour(x, y, z, colors='blue', levels=3, zorder=2)

    @staticmethod
    def remove_collection(collection):
        if isinstance(collection, QuadContourSet):
            iter_obj = collection.collections
        else:
            iter_obj = collection

        for artist in iter_obj:
            artist.remove()

    def proposed_points(self, ax, points):
        two_dim_points = np.atleast_2d(points)
        return ax.scatter(two_dim_points[:, 0], two_dim_points[:, 1], s=self.proposal_size, color='orange', zorder=3)

    @staticmethod
    def arrows(ax, start_points, end_points, color='black', alpha=1.):
        out_arrows = []
        two_dim_start_points = np.atleast_2d(start_points)
        two_dim_end_points = np.atleast_2d(end_points)

        if isinstance(alpha, float):
            alpha = np.ones(len(two_dim_start_points)) * alpha

        for i in range(len(two_dim_start_points)):
            if not np.array_equal(two_dim_start_points[i], two_dim_end_points[i]):
                out_arrows.append(ax.annotate("", xy=two_dim_end_points[i], xytext=two_dim_start_points[i],
                                              arrowprops=dict(arrowstyle='-|>', color=color, alpha=float(alpha[i])),
                                              zorder=2))
        return out_arrows

    @staticmethod
    def leapfrog_points(ax, x_points, alpha=1):
        return ax.plot(x_points[:, 0], x_points[:, 1], marker='.', color='black', zorder=2.5, alpha=alpha)

    @staticmethod
    def arrayify(cdict):
        out_cdict = cdict.copy()
        for key, value in out_cdict.__dict__.items():
            if isinstance(value, (float, int)):
                out_cdict.__dict__[key] = np.asarray(value)

        return out_cdict

    @staticmethod
    def vis_run_mcmc(scenario,
                     sampler,
                     n,
                     correction):

        initial_state = RunVisUtils.arrayify(sampler.initial_state)
        initial_extra = RunVisUtils.arrayify(sampler.initial_extra)

        @jit
        def markov_kernel(previous_carry: Tuple[CDict, CDict],
                          _: Any) -> Tuple[Tuple[CDict, CDict], Tuple[CDict, CDict, CDict]]:
            previous_state, previous_extra = previous_carry
            reject_state = previous_state.copy()
            reject_extra = previous_extra.copy()
            reject_extra.iter += 1

            reject_state, reject_extra = sampler.always(scenario,
                                                        reject_state,
                                                        reject_extra)

            proposed_state, proposed_extra = sampler.proposal(scenario,
                                                              reject_state,
                                                              reject_extra)

            corrected_state, corrected_extra = correction(scenario, sampler,
                                                          reject_state, reject_extra,
                                                          proposed_state, proposed_extra)

            return (corrected_state, corrected_extra), (corrected_state, corrected_extra, proposed_state)

        # Run Sampler
        start = time()
        final_carry, chain = scan(markov_kernel,
                                  (sampler.initial_state, sampler.initial_extra),
                                  None,
                                  length=n)
        end = time()

        corrected_samples, corrected_extras, proposed_samples = chain

        corrected_samples = initial_state[np.newaxis] + corrected_samples
        corrected_extras = initial_extra[np.newaxis] + corrected_extras

        if proposed_samples.value.shape[1:] == initial_state.value.shape:
            proposed_samples = initial_state[np.newaxis] + proposed_samples
        elif proposed_samples.value.shape[1:] == initial_state.value.shape[1:]:
            proposed_samples = initial_state[np.newaxis, 0] + proposed_samples
        else:
            raise ValueError("Couldn't combine initial_state and proposed_samples")

        corrected_samples.time = end - start

        return corrected_samples, corrected_extras, proposed_samples


class RunVis:

    def __init__(self,
                 ax: plt.Axes,
                 scenario: TwoDimScenario,
                 sampler: Union[Sampler, MCMCSampler],
                 correction: Correction,
                 n: int,
                 utils: RunVisUtils = RunVisUtils()):
        self.utils = utils
        self.ax = ax
        self.scenario = scenario.copy()
        self.sampler = sampler
        self.correction = correction
        self.n = n

        self.leapfrog = hasattr(sampler.initial_state, 'momenta') and hasattr(sampler.parameters, 'leapfrog_steps')
        self.requires_gradient = hasattr(sampler.initial_state, 'grad_potential')
        self.ensemble = self.sampler.initial_state.value.ndim == 2

        if self.leapfrog:
            self.sampler.initial_state._all_leapfrog_value = np.zeros((self.sampler.parameters.leapfrog_steps + 1,
                                                                       self.sampler.initial_state.value.shape[-1]))
            self.sampler.initial_state._all_leapfrog_momenta = np.zeros((2 * self.sampler.parameters.leapfrog_steps + 1,
                                                                         *self.sampler.initial_state.momenta.shape))

        self.run_params = mcmc_run_params(sampler, correction)
        self.corrected_samples, self.corrected_extras, self.proposed_samples = self.utils.vis_run_mcmc(self.scenario,
                                                                                                       self.sampler,
                                                                                                       n,
                                                                                                       self.correction)

        self.reject_state = self.corrected_samples[0]
        self.reject_extra = self.corrected_extras[1]
        self.proposed_state = self.proposed_samples[1]
        self.corrected_state = self.corrected_samples[1]

        self.plot_space = None
        self.frames_per_sample = None
        self.live_frame_index = 0
        self.sample_index = 0
        self.scenario_dens = None
        self.full_state_points = None
        self.live_state_points = None

        self.orig_xlim = None
        self.orig_ylim = None

    def anim_init(self):
        self.sample_index = 1
        self.live_frame_index = 0

        self.scenario.auto_axes_lims()
        self.ax.set_xlim(self.scenario.xlim)
        self.ax.set_ylim(self.scenario.ylim)
        self.scenario_dens = self.utils.scenario_contours(self.ax, self.scenario, self.utils.plot_scen_potential)
        self.plot_space = _generate_plot_grid(self.scenario.xlim, self.scenario.ylim,
                                              self.scenario.plot_resolution, True)
        self.full_state_points = self.utils.full_state_points(self.ax, self.reject_state.value[np.newaxis])
        self.live_state_points = self.utils.live_state_points(self.ax, self.reject_state.value)
        self.ax.autoscale(True)
        self.check_scenario_lims()
        self.orig_xlim = self.ax.get_xlim()
        self.orig_ylim = self.ax.get_ylim()

    def __call__(self, i):

        self.run_vis()
        self.check_scenario_lims()

        if self.live_frame_index == (self.frames_per_sample - 1):

            self.live_state_points.set_offsets(self.corrected_state.value)

            self.check_deexpand_lims()
            self.sample_index += 1
            self.reject_state = self.corrected_samples[self.sample_index - 1]
            self.reject_extra = self.corrected_extras[self.sample_index]
            self.proposed_state = self.proposed_samples[self.sample_index]
            self.corrected_state = self.corrected_samples[self.sample_index]
            
            if self.ensemble:
                self.full_state_points.set_offsets(np.concatenate(self.corrected_samples.value[:self.sample_index]))
            else:
                self.full_state_points.set_offsets(self.corrected_samples.value[:self.sample_index])

        self.live_frame_index = (self.live_frame_index + 1) % self.frames_per_sample

    def check_deexpand_lims(self):
        # Check if rejected proposal altered limits
        ax_xlim = self.ax.get_xlim()
        ax_ylim = self.ax.get_ylim()

        if ax_xlim != self.orig_xlim or ax_ylim != self.orig_ylim:

            left_and_returned = np.min(self.corrected_state.value[..., 0]) > self.orig_xlim[0] \
                                and np.max(self.corrected_state.value[..., 0]) < self.orig_xlim[1] \
                                and np.min(self.corrected_state.value[..., 1]) > self.orig_ylim[0] \
                                and np.max(self.corrected_state.value[..., 1]) < self.orig_ylim[1]

            if left_and_returned:
                self.ax.set_xlim(*self.orig_xlim, auto=True)
                self.ax.set_ylim(*self.orig_ylim, auto=True)
                self.ax.relim()

        self.check_scenario_lims()
        self.orig_xlim = self.ax.get_xlim()
        self.orig_ylim = self.ax.get_ylim()

    def check_scenario_lims(self):
        ax_xlim = self.ax.get_xlim()
        ax_ylim = self.ax.get_ylim()

        if ax_xlim != self.scenario.xlim or ax_ylim != self.scenario.ylim:
            self.scenario.xlim = ax_xlim
            self.scenario.ylim = ax_ylim
            self.utils.remove_collection(self.scenario_dens)
            self.scenario_dens = self.utils.scenario_contours(self.ax, self.scenario, self.utils.plot_scen_potential)
            self.plot_space = _generate_plot_grid(self.scenario.xlim, self.scenario.ylim,
                                                  self.scenario.plot_resolution, True)

    def run_vis(self):
        raise NotImplementedError


class UncorrectedRunVis(RunVis):

    def __init__(self,
                 ax: plt.Axes,
                 scenario: TwoDimScenario,
                 sampler: MCMCSampler,
                 correction: Correction,
                 n: int,
                 utils: RunVisUtils = RunVisUtils()):
        super().__init__(ax, scenario, sampler, correction, n, utils)
        self.frames_per_sample = 2 + self.sampler.parameters.leapfrog_steps if self.leapfrog else 2

        try:
            prop_pot = self.sampler.proposal_potential(self.scenario,
                                                       self.corrected_samples[0], self.corrected_extras[0],
                                                       CDict(value=np.zeros((3, 3, 2))), self.corrected_extras[1])

            if prop_pot.shape == (3, 3):
                self.proposal_potential_available = True
            else:
                self.proposal_potential_available = False

        except AttributeError:
            self.proposal_potential_available = False

    def update_proposed_points(self):
        if self.proposal_potential_available:
            z_proposal_potential = self.sampler.proposal_potential(self.scenario,
                                                                   self.reject_state, self.reject_extra,
                                                                   CDict(value=self.plot_space[2]), self.reject_extra)

            z_proposal_dens = np.exp(-z_proposal_potential)
            self.proposal_dens_contours = self.utils.proposal_contours(self.ax,
                                                                       self.plot_space[0],
                                                                       self.plot_space[1],
                                                                       z_proposal_dens)

        self.proposed_points = self.utils.proposed_points(self.ax, self.proposed_state.value)

        if not self.leapfrog and not hasattr(self.proposed_state, 'momenta'):
            self.arrows = self.utils.arrows(self.ax, self.reject_state.value, self.proposed_state.value)
            if self.requires_gradient:
                if self.proposed_state.value.ndim == 2:
                    alphas = np.where(np.arange(self.sampler.parameters.n_ensemble)
                                      == self.reject_extra.iter % self.sampler.parameters.n_ensemble,
                                      self.utils.gradient_alpha, 0.)
                else:
                    alphas = self.utils.gradient_alpha

                self.arrows = self.arrows + self.utils.arrows(self.ax, self.reject_state.value,
                                                              self.reject_state.value
                                                              - self.reject_state.grad_potential,
                                                              alpha=alphas)

    def clear_live_points(self):
        if self.leapfrog:
            self.utils.remove_collection(self.leapfrog_points)
        else:
            self.utils.remove_collection(self.arrows)
        self.proposed_points.remove()

        if self.proposal_potential_available:
            self.utils.remove_collection(self.proposal_dens_contours)

    def update_leapfrog_points(self):
        leapfrog_x = self.proposed_state._all_leapfrog_value
        leapfrog_p = self.proposed_state._all_leapfrog_momenta

        if self.live_frame_index > 0:
            self.utils.remove_collection(self.leapfrog_points)
        self.leapfrog_points = self.utils.leapfrog_points(self.ax, leapfrog_x[:self.live_frame_index + 2])

        if hasattr(self.proposed_state, 'momenta') and self.proposed_state.value.ndim == 2:
            if hasattr(self, 'arrows'):
                self.utils.remove_collection(self.arrows)

            alphas = np.where(np.arange(self.sampler.parameters.n_ensemble)
                              == self.reject_extra.iter % self.sampler.parameters.n_ensemble,
                              0., self.utils.gradient_alpha)

            self.arrows = self.utils.arrows(self.ax,
                                            self.proposed_state.value,
                                            self.proposed_state.value
                                            + self.reject_extra.parameters.stepsize * leapfrog_p[
                                                2 * self.live_frame_index + 2],
                                            alpha=alphas)

        self.check_scenario_lims()

    def run_vis(self):

        if self.live_frame_index == self.frames_per_sample - 2:
            self.update_proposed_points()

        elif self.live_frame_index == self.frames_per_sample - 1:
            self.clear_live_points()

        elif self.live_frame_index >= self.frames_per_sample:
            raise IndexError("Visualisation live_frame_index exceeded frames_per_sample")

        else:
            self.update_leapfrog_points()


class MHRunVis(UncorrectedRunVis):

    def __init__(self,
                 ax: plt.Axes,
                 scenario: TwoDimScenario,
                 sampler: MCMCSampler,
                 correction: Correction,
                 n: int,
                 utils: RunVisUtils = RunVisUtils()):
        super().__init__(ax, scenario, sampler, correction, n, utils)
        self.frames_per_sample = 3 + self.sampler.parameters.leapfrog_steps if self.leapfrog else 3

    def run_vis(self):

        if self.live_frame_index == self.frames_per_sample - 3:
            super().update_proposed_points()

        elif self.live_frame_index == self.frames_per_sample - 2:
            if np.array_equal(self.proposed_state.value, self.corrected_state.value):
                self.proposed_points.set_color('green')
            else:
                self.proposed_points.set_color('red')

        elif self.live_frame_index == self.frames_per_sample - 1:
            super().clear_live_points()

        elif self.live_frame_index >= self.frames_per_sample:
            raise IndexError("Visualisation live_frame_index exceeded frames_per_sample")

        else:
            super().update_leapfrog_points()


def visualise(scenario: TwoDimScenario,
              sampler: MCMCSampler,
              random_key: np.ndarray,
              correction: Union[None, str, Correction, Type[Correction]] = 'sampler_default',
              run_vis: Type[RunVis] = None,
              n: int = 100,
              ms_per_sample: float = 2000,
              potential: bool = False,
              return_sample: bool = False,
              utils: RunVisUtils = RunVisUtils(),
              **kwargs):
    sampler, correction = startup_mcmc(scenario, sampler, random_key, correction, **kwargs)

    utils.plot_scen_potential = potential

    fig, ax = plt.subplots()
    plt.tight_layout()

    if run_vis is None:
        if isinstance(correction, Uncorrected):
            run_vis = UncorrectedRunVis(ax, scenario, sampler, correction, n, utils)
        elif isinstance(correction, Metropolis):
            run_vis = MHRunVis(ax, scenario, sampler, correction, n, utils)
        else:
            raise ValueError(f'RunVis not found for correction: {correction}')
    elif isclass(run_vis) and issubclass(run_vis, RunVis):
        run_vis = run_vis(ax, scenario, sampler, correction, n, utils)
    elif isinstance(run_vis, RunVis):
        raise ValueError(f'run_vis {run_vis} given as instance rather than (uninitiated) class')
    else:
        raise ValueError(f'run_vis {run_vis} not understood')

    anim = animation.FuncAnimation(fig,
                                   run_vis,
                                   frames=(n - 1) * run_vis.frames_per_sample,
                                   init_func=run_vis.anim_init,
                                   interval=ms_per_sample / run_vis.frames_per_sample,
                                   repeat=False)

    if return_sample:
        out_sample = run_vis.corrected_samples
        out_sample.run_params = run_vis.run_params
        return anim, out_sample
    else:
        return anim



########################################################################################################################
# Module: mcmc/ensemble_mcmc.py
# Description: MCMC which uses an ensemble to adapt the proposal covariance.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Tuple, Union

import jax.numpy as np
from jax import random, vmap
from jax.lax import cond
from jax.ops import index_update

from mocat.src.mcmc.sampler import MCMCSampler, default_initial_extra
from mocat.src import utils
from mocat.src.core import Scenario, CDict
from mocat.src.mcmc.corrections import Metropolis


# Ensemble Random Walk with sample covariance preconditioner
class EnsembleRWMH(MCMCSampler):
    name = 'Ensemble Random Walk'
    default_correction = Metropolis()

    def __init__(self,
                 stepsize: float = None,
                 n_ensemble: int = None,
                 identity_scaling: float = 0.1,
                 samp_cov_scaling: float = 1.):
        super().__init__()
        self.parameters.stepsize = stepsize
        self.parameters.n_ensemble = n_ensemble
        self.parameters.identity_scaling = identity_scaling
        self.parameters.samp_cov_scaling = samp_cov_scaling
        self.tuning.target = 0.234

    def startup(self,
                scenario: Scenario,
                random_key: np.ndarray):
        if self.initial_state is None \
                or not hasattr(self.initial_state, 'value') \
                or self.initial_state.value.shape[-1] != scenario.dim \
                or self.initial_state.value.shape[1] != self.parameters.n_ensemble:
            random_key, sub_key = random.split(random_key)

            x0 = random.normal(sub_key, shape=(self.parameters.n_ensemble, scenario.dim))
            self.initial_state = CDict(value=x0)
        default_initial_extra(self, random_key)

    def proposal(self,
                 scenario: Scenario,
                 reject_state: CDict,
                 reject_extra: CDict) -> Tuple[CDict, CDict]:
        proposed_state = reject_state.copy()

        d = scenario.dim
        n_ensemble = self.parameters.n_ensemble
        ensemble = reject_state.value
        ensemble_index = reject_extra.iter % n_ensemble

        stepsize = reject_extra.parameters.stepsize
        identity_scaling = reject_extra.parameters.identity_scaling
        samp_cov_scaling = reject_extra.parameters.samp_cov_scaling


        leave_one_out_ensemble = np.where(np.expand_dims(np.arange(n_ensemble - 1), 1) < ensemble_index,
                                          ensemble[:-1],
                                          ensemble[1:])
        leave_one_out_cov = samp_cov_scaling * np.atleast_2d(np.cov(leave_one_out_ensemble.T))\
                            + identity_scaling * np.eye(d)
        leave_one_out_cov_sqrt = np.linalg.cholesky(leave_one_out_cov)

        reject_extra.random_key, subkey = random.split(reject_extra.random_key)

        precon_noise = leave_one_out_cov_sqrt @ random.normal(subkey, (d,))

        proposed_state.value = index_update(proposed_state.value, ensemble_index,
                                            reject_state.value[ensemble_index] + np.sqrt(stepsize) * precon_noise)

        return proposed_state, reject_extra

    def acceptance_probability(self,
                               scenario: Scenario,
                               reject_state: CDict, reject_extra: CDict,
                               proposed_state: CDict, proposed_extra: CDict) -> Union[float, np.ndarray]:
        ensemble_index = reject_extra.iter % self.parameters.n_ensemble
        return np.minimum(1., np.exp(- proposed_state.potential[ensemble_index]
                                     + reject_state.potential[ensemble_index]))


# Ensemble Overdamped Langevin with sample covariance preconditioner
class EnsembleOverdamped(MCMCSampler):
    name = 'Ensemble Overdamped'
    default_correction = Metropolis()

    def __init__(self,
                 stepsize: float = None,
                 n_ensemble: int = None,
                 identity_scaling: float = 0.1,
                 samp_cov_scaling: float = 1.):
        super().__init__()
        self.parameters.stepsize = stepsize
        self.parameters.n_ensemble = n_ensemble
        self.parameters.identity_scaling = identity_scaling
        self.parameters.samp_cov_scaling = samp_cov_scaling
        self.tuning.target = 0.574

    def startup(self,
                scenario: Scenario,
                random_key: np.ndarray):
        if self.initial_state is None \
                or not hasattr(self.initial_state, 'value') \
                or self.initial_state.value.shape[-1] != scenario.dim \
                or self.initial_state.value.shape[1] != self.parameters.n_ensemble:
            random_key, sub_key = random.split(random_key)

            x0 = random.normal(sub_key, shape=(self.parameters.n_ensemble, scenario.dim))
            self.initial_state = CDict(value=x0)
        self.initial_state.grad_potential = vmap(scenario.grad_potential)(self.initial_state.value)
        default_initial_extra(self, random_key)
        self.initial_state, self.initial_extra = self.always(scenario, self.initial_state, self.initial_extra)

    def always(self,
               scenario: Scenario,
               reject_state: CDict,
               reject_extra: CDict) -> Tuple[CDict, CDict]:
        d = scenario.dim
        n_ensemble = self.parameters.n_ensemble
        ensemble = reject_state.value
        ensemble_index = reject_extra.iter % n_ensemble

        identity_scaling = reject_extra.parameters.identity_scaling
        samp_cov_scaling = reject_extra.parameters.samp_cov_scaling

        leave_one_out_ensemble = np.where(np.expand_dims(np.arange(n_ensemble - 1), 1) < ensemble_index,
                                          ensemble[:-1],
                                          ensemble[1:])
        leave_one_out_cov = samp_cov_scaling * np.atleast_2d(
            np.cov(leave_one_out_ensemble.T)) + identity_scaling * np.eye(d)
        reject_extra.leave_one_out_cov = leave_one_out_cov
        reject_extra.leave_one_out_prec = np.linalg.inv(leave_one_out_cov)
        reject_extra.leave_one_out_cov_sqrt = np.linalg.cholesky(leave_one_out_cov)
        return reject_state, reject_extra

    def proposal(self,
                 scenario: Scenario,
                 reject_state: CDict,
                 reject_extra: CDict) -> Tuple[CDict, CDict]:
        proposed_state = reject_state.copy()

        d = scenario.dim
        n_ensemble = self.parameters.n_ensemble
        ensemble_index = reject_extra.iter % n_ensemble

        stepsize = reject_extra.parameters.stepsize

        leave_one_out_cov = reject_extra.leave_one_out_cov
        leave_one_out_cov_sqrt = reject_extra.leave_one_out_cov_sqrt

        reject_extra.random_key, subkey = random.split(reject_extra.random_key)

        precon_noise = leave_one_out_cov_sqrt @ random.normal(subkey, (d,))

        proposed_state.value = index_update(reject_state.value, ensemble_index,
                                            reject_state.value[ensemble_index]
                                            - stepsize * leave_one_out_cov @ reject_state.grad_potential[ensemble_index]
                                            + np.sqrt(2 * stepsize) * precon_noise)

        proposed_state.grad_potential = index_update(proposed_state.grad_potential, ensemble_index,
                                                     scenario.grad_potential(proposed_state.value[ensemble_index]))

        reject_extra.leave_one_out_cov = leave_one_out_cov

        return proposed_state, reject_extra

    def proposal_potential(self,
                           scenario: Scenario,
                           reject_state: CDict, reject_extra: CDict,
                           proposed_state: CDict, proposed_extra: CDict) -> Union[float, np.ndarray]:
        ensemble_index = reject_extra.iter % self.parameters.n_ensemble
        stepsize = reject_extra.parameters.stepsize
        leave_one_out_cov = reject_extra.leave_one_out_cov
        leave_one_out_prec = reject_extra.leave_one_out_prec

        if proposed_state.value.ndim == 2:
            prop_val = proposed_state.value[ensemble_index]
        else:
            prop_val = proposed_state.value

        return utils.gaussian_potential(prop_val,
                                        reject_state.value[ensemble_index]
                                        - stepsize * leave_one_out_cov @
                                        reject_state.grad_potential[ensemble_index],
                                        leave_one_out_prec / (2 * stepsize))

    def acceptance_probability(self,
                               scenario: Scenario,
                               reject_state: CDict, reject_extra: CDict,
                               proposed_state: CDict, proposed_extra: CDict) -> Union[float, np.ndarray]:
        ensemble_index = reject_extra.iter % self.parameters.n_ensemble

        pre_min_alpha = np.exp(- proposed_state.potential[ensemble_index]
                               + reject_state.potential[ensemble_index]
                               - self.proposal_potential(scenario,
                                                         proposed_state, proposed_extra,
                                                         reject_state, reject_extra)
                               + self.proposal_potential(scenario,
                                                         reject_state, reject_extra,
                                                         proposed_state, proposed_extra))

        return np.minimum(1., pre_min_alpha)



########################################################################################################################
# Module: scenarios/twodim/toy_scenarios.py
# Description: Initiates some toy 2D scenarios suitable for visualisation.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Union

import jax.numpy as np

from mocat.src.scenarios.twodim.vectorise import TwoDimScenario
from mocat.src.scenarios.toy_scenarios import Gaussian as NDGaussian
from mocat.src.scenarios.toy_scenarios import GaussianMixture as NDGaussianMixture
from mocat.src.scenarios.toy_scenarios import DoubleWell as NDDoubleWell
from mocat.src.scenarios.toy_scenarios import Rastrigin as NDRastrigin


class Gaussian(TwoDimScenario, NDGaussian):
    def __init__(self,
                 mean: np.ndarray = np.array([0., 0.]),
                 cov: np.ndarray = None):
        TwoDimScenario.__init__(self)
        NDGaussian.__init__(self, 2, mean, cov)


class Banana(TwoDimScenario):
    name = 'Banana'
    curviness = 0.03
    lengthiness = 100

    def potential(self, x: np.ndarray) -> Union[float, np.ndarray]:
        return 0.5 * (x[0] ** 2 / self.lengthiness + (x[1] + self.curviness * x[0] ** 2
                                                      - self.lengthiness * self.curviness) ** 2)


class NealFunnel(TwoDimScenario):
    name = "Neal's Funnel"

    def potential(self, x: np.ndarray) -> Union[float, np.ndarray]:
        return 0.5 * (x[1] ** 2 / 3 + x[0] ** 2 * np.exp(-x[1] / 2))


class GaussianMixture(TwoDimScenario, NDGaussianMixture):
    dim = 2

    def __init__(self,
                 means: np.ndarray = np.array([[-2, -2], [2, -2], [-2, 2], [2, 2]]),
                 **kwargs):
        TwoDimScenario.__init__(self)
        NDGaussianMixture.__init__(self, means, **kwargs)


class DoubleWell(TwoDimScenario, NDDoubleWell):
    def __init__(self):
        TwoDimScenario.__init__(self)
        NDDoubleWell.__init__(self, dim=2)


class Rastrigin(TwoDimScenario, NDRastrigin):
    def __init__(self,
                 a: float = 1.):
        TwoDimScenario.__init__(self)
        NDRastrigin.__init__(self, dim=2, a=a)





########################################################################################################################
# Module: scenarios/toy_scenarios.py
# Description: Some simple multivariate scenarios
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Union

import jax.numpy as np
from jax import vmap, jit

from mocat.src.core import Scenario


class Gaussian(Scenario):
    name = "Gaussian"

    def __init__(self,
                 dim: int = 1,
                 mean: Union[float, np.ndarray] = None,
                 cov: Union[float, np.ndarray] = None):
        if mean is not None:
            self.dim = mean.shape[-1]
        elif cov is not None:
            self.dim = cov.shape[-1]
        else:
            self.dim = dim
        self.mean = np.zeros(dim) if mean is None else mean
        self.cov = np.eye(dim) if cov is None else cov
        super().__init__()

    def potential(self, x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        x_diff = (x - self.mean) @ self.sqrt_prec
        return 0.5 * np.sum(np.square(x_diff), axis=-1)

    def __setattr__(self, key, value):
        self.__dict__[key] = value
        if key == 'cov':
            self.prec = np.eye(self.dim) if value is None else np.linalg.inv(value)
            self.sqrt_prec = np.eye(self.dim) if value is None else np.linalg.cholesky(self.prec)


class GaussianMixture(Scenario):
    name = "Gaussian Mixture"

    def __init__(self,
                 means: np.ndarray,
                 covs: np.ndarray = None,
                 weights: np.ndarray = None):
        self.means = means[:, np.newaxis] if means.ndim == 1 else means
        self.dim = self.means.shape[-1]
        if weights is None:
            # Default: equal mixture weights
            self.weights = np.ones(self.n_components) / self.n_components
        else:
            self.weights = weights
        self.covs = covs

        super().__init__()

    def __setattr__(self, key, value):
        self.__dict__[key] = value
        if key == 'covs':
            if value is None:
                # Default: identity covariances
                self.covs = np.array([np.eye(self.dim)] * self.n_components)
            elif value.shape == (self.dim, self.dim):
                self.covs = np.repeat(value[np.newaxis, :, :], self.n_components, axis=0)

            self.precs = vmap(np.linalg.inv)(self.covs)
            self.sqrt_precs = vmap(np.linalg.cholesky)(self.precs)
            self.det_precs = vmap(np.linalg.det)(self.precs)

            self.component_potential = jit(self.component_potential)
            self.component_potentials = jit(vmap(self.component_potential, (None, 0), 0))
            self.component_dens = jit(self.component_dens)
            self.dens = jit(self.dens)
            self.component_potential = jit(self.component_potential)

    @property
    def n_components(self) -> int:
        return len(self.means)

    def component_potential(self,
                            x: np.ndarray,
                            component_index: int) -> Union[float, np.ndarray]:
        return 0.5 * np.sum(np.square((x - self.means[component_index]) @
                                      self.sqrt_precs[component_index]), axis=-1) \
               - np.log(self.weights[component_index]
                        * self.det_precs[component_index]
                        / np.power(2 * np.pi, self.dim * 0.5))

    def component_dens(self,
                       x: np.ndarray,
                       component_index: int) -> Union[float, np.ndarray]:
        return np.exp(-self.component_potential(x, component_index))

    def dens(self, x: np.ndarray) -> Union[float, np.ndarray]:
        return np.sum(np.exp(-self.component_potentials(x, np.arange(self.n_components))), axis=0)

    def potential(self, x: np.ndarray) -> Union[float, np.ndarray]:
        return -np.log(self.dens(x))


class DoubleWell(Scenario):
    name = "Double Well"

    def __init__(self,
                 dim: int = 1):
        self.dim = dim
        super().__init__()

    def potential(self, x: np.ndarray) -> Union[float, np.ndarray]:
        return np.sum(np.power(x, 4), axis=-1) / 4. \
               - np.sum(np.power(x, 2), axis=-1) / 2.


class Rastrigin(Scenario):
    name = "Rastrigin"

    def __init__(self,
                 dim: int = 1,
                 a: float = 1.):
        self.dim = dim
        self.a = a
        super().__init__()

    def potential(self, x: np.ndarray) -> Union[float, np.ndarray]:
        return self.a*self.dim + np.sum(x**2 - self.a * np.cos(2 * np.pi * x), axis=-1)



########################################################################################################################
# Module: tests/test_toy_scenarios.py
# Description: Tests for univariate and multivariate toy scenarios
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################


import unittest

import jax.numpy as np
import numpy.testing as npt

from mocat.src.scenarios import toy_scenarios


class Test1DGaussian(unittest.TestCase):

    scenario = toy_scenarios.Gaussian(1)

    def test_basic(self):
        npt.assert_equal(self.scenario.dim, 1)
        npt.assert_equal(self.scenario.mean.shape, (1,))
        npt.assert_equal(self.scenario.cov.shape, (1, 1))
        npt.assert_equal(self.scenario.prec.shape, (1, 1))
        npt.assert_equal(self.scenario.sqrt_prec.shape, (1, 1))

    def test_scalar_potential(self):
        npt.assert_array_equal(self.scenario.potential(0.), 0.)
        npt.assert_array_equal(self.scenario.potential(5.), 12.5)

    def test_scalar_grad_potential(self):
        npt.assert_array_equal(self.scenario.grad_potential(0.), 0.)
        npt.assert_array_equal(self.scenario.grad_potential(5.), 5.)

    def test_setcov(self):
        self.scenario.cov = np.array([[7.]])
        npt.assert_array_equal(self.scenario.prec, 1/7.)
        npt.assert_array_almost_equal(self.scenario.sqrt_prec, 0.37796447)


class TestNDGaussian(unittest.TestCase):

    scenario = toy_scenarios.Gaussian(10)

    def test_basic(self):
        npt.assert_equal(self.scenario.dim, 10)
        npt.assert_equal(self.scenario.mean.shape, (10,))
        npt.assert_equal(self.scenario.cov.shape, (10, 10))
        npt.assert_equal(self.scenario.prec.shape, (10, 10))
        npt.assert_equal(self.scenario.sqrt_prec.shape, (10, 10))

    def test_array_potential(self):
        npt.assert_array_equal(self.scenario.potential(np.zeros(10)), 0.)
        npt.assert_array_equal(self.scenario.potential(np.ones(10) * 3), 45.)

    def test_array_grad_potential(self):
        npt.assert_array_equal(self.scenario.grad_potential(np.zeros(10)), np.zeros(10))
        npt.assert_array_equal(self.scenario.grad_potential(np.ones(10) * 3.), np.ones(10) * 3.)


class Test1DGaussianMixture(unittest.TestCase):

    scenario = toy_scenarios.GaussianMixture(means=np.array([0, 1]))

    def test_basic(self):
        npt.assert_equal(self.scenario.dim, 1)
        npt.assert_equal(self.scenario.means.shape, (2, 1))
        npt.assert_equal(self.scenario.covs.shape, (2, 1, 1))
        npt.assert_equal(self.scenario.precs.shape, (2, 1, 1))
        npt.assert_equal(self.scenario.sqrt_precs.shape, (2, 1, 1))
        npt.assert_equal(self.scenario.det_precs.shape, (2,))

    def test_scalar_potential(self):
        npt.assert_array_equal(self.scenario.potential(0.), self.scenario.potential(1.))
        npt.assert_array_almost_equal(self.scenario.potential(5.), 9.601038)

    def test_scalar_grad_potential(self):
        npt.assert_array_equal(self.scenario.grad_potential(0.), -self.scenario.grad_potential(1.))
        npt.assert_array_almost_equal(self.scenario.grad_potential(5.), 4.010987)

    def test_setcovs(self):
        self.scenario.covs = np.repeat(np.array([[7.]])[np.newaxis, :, :], 2, axis=0)
        npt.assert_array_equal(self.scenario.precs, np.repeat(np.array([[1/7.]])[np.newaxis, :, :], 2, axis=0))
        npt.assert_array_almost_equal(self.scenario.sqrt_precs,
                                      np.repeat(np.array([[0.37796447]])[np.newaxis, :, :], 2, axis=0))


class TestNDGaussianMixture(unittest.TestCase):

    scenario = toy_scenarios.GaussianMixture(means=np.arange(15).reshape(5, 3))

    def test_basic(self):
        npt.assert_equal(self.scenario.dim, 3)
        npt.assert_equal(self.scenario.means.shape, (5, 3))
        npt.assert_equal(self.scenario.covs.shape, (5, 3, 3))
        npt.assert_equal(self.scenario.precs.shape, (5, 3, 3))
        npt.assert_equal(self.scenario.sqrt_precs.shape, (5, 3, 3))
        npt.assert_equal(self.scenario.det_precs.shape, (5,))

    def test_array_potential(self):
        npt.assert_array_almost_equal(self.scenario.potential(np.zeros(3)), 6.8662534)
        npt.assert_array_almost_equal(self.scenario.potential(np.ones(3) * 2), 6.8552055)

    def test_array_grad_potential(self):
        npt.assert_array_almost_equal(self.scenario.grad_potential(np.zeros(3)), np.array([0., -1, -2]))
        npt.assert_array_almost_equal(self.scenario.grad_potential(np.ones(3) * 2.), np.array([1.967039,
                                                                                               0.967039,
                                                                                               -0.032961]))


class Test1DDoubleWell(unittest.TestCase):

    scenario = toy_scenarios.DoubleWell(1)

    def test_basic(self):
        npt.assert_equal(self.scenario.dim, 1)

    def test_scalar_potential(self):
        npt.assert_array_equal(self.scenario.potential(-3.), self.scenario.potential(3.))
        npt.assert_array_equal(self.scenario.potential(0.), 0)
        npt.assert_array_almost_equal(self.scenario.potential(5.), 143.75)

    def test_scalar_grad_potential(self):
        npt.assert_array_equal(self.scenario.grad_potential(-3.), -self.scenario.grad_potential(3.))
        npt.assert_array_equal(self.scenario.grad_potential(0.), 0.)
        npt.assert_array_equal(self.scenario.grad_potential(1.), 0.)
        npt.assert_array_equal(self.scenario.grad_potential(-1.), 0.)
        npt.assert_array_almost_equal(self.scenario.grad_potential(5.), 120.)


class TestNDDoubleWell(unittest.TestCase):

    scenario = toy_scenarios.DoubleWell(6)

    def test_basic(self):
        npt.assert_equal(self.scenario.dim, 6)

    def test_array_potential(self):
        npt.assert_array_equal(self.scenario.potential(np.zeros(6)), 0.)
        npt.assert_array_equal(self.scenario.potential(np.ones(3) * 2), 6.)
        npt.assert_array_almost_equal(self.scenario.potential(np.arange(6)), 217.25)

    def test_array_grad_potential(self):
        npt.assert_array_equal(self.scenario.grad_potential(np.zeros(6)), np.zeros(6))
        npt.assert_array_equal(self.scenario.grad_potential(np.ones(6)), np.zeros(6))
        npt.assert_array_equal(self.scenario.grad_potential(-np.zeros(6)), np.zeros(6))
        npt.assert_array_almost_equal(self.scenario.grad_potential(np.arange(6, dtype='float32')),
                                      np.array([0., 0., 6., 24., 60., 120.]))


if __name__ == '__main__':
    unittest.main()



########################################################################################################################
# Module: utils.py
# Description: Some useful functions including Gaussian potential evaluations and leapfrog integrator.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Any, Union, Callable, Tuple
from functools import partial

from decorator import decorator
import jax.numpy as np
from jax.lax import scan
from jax import jit
import matplotlib.pyplot as plt
from matplotlib.figure import Figure

from mocat.src.core import CDict


@partial(jit, static_argnums=(0,))
def leave_one_out_indices(n: int,
                          remove_index: int) -> np.ndarray:
    full_inds = np.arange(n + 1)
    leave_one_out_inds = np.where(np.arange(n) < remove_index, full_inds[:-1], full_inds[1:])
    return leave_one_out_inds


@jit
def _vectorised_gaussian_potential(x: np.ndarray,
                                   mean: Union[np.ndarray, float],
                                   sqrt_prec: np.ndarray) -> Union[np.ndarray, float]:
    return 0.5 * np.sum(np.square((x - mean) @ sqrt_prec), axis=-1)


@jit
def _vectorised_gaussian_potential_identcov(x: np.ndarray,
                                            mean: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
    return 0.5 * np.sum(np.square(x - mean), axis=-1)


@jit
def _mv_gaussian_potential(x: np.ndarray,
                           mean: Union[np.ndarray, float],
                           prec: np.ndarray) -> Union[np.ndarray, float]:
    diff = x - mean
    return 0.5 * diff.T @ prec @ diff


@jit
def _mv_gaussian_potential_identcov(x: np.ndarray,
                                    mean: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
    diff = x - mean
    return 0.5 * np.sum(np.square(diff), axis=-1)


def gaussian_potential(x: np.ndarray,
                       mean: Union[np.ndarray, float] = 0.,
                       prec: np.ndarray = None,
                       sqrt_prec: np.ndarray = None) -> np.ndarray:
    if x.ndim == 1:
        if prec is None:
            return _mv_gaussian_potential_identcov(x, mean)
        else:
            return _mv_gaussian_potential(x, mean, prec)
    else:
        if prec is not None and sqrt_prec is None:
            sqrt_prec = np.linalg.cholesky(prec)

        if sqrt_prec is None:
            return _vectorised_gaussian_potential_identcov(x, mean)
        else:
            return _vectorised_gaussian_potential(x, mean, sqrt_prec)


@partial(jit, static_argnums=(1, 3))
def leapfrog(previous_state: CDict,
             grad_potential: Callable,
             stepsize: float,
             n_steps: int) -> CDict:
    def leapfrog_step(init_state: CDict,
                      _: Any):
        new_state = init_state.copy()

        p_half = init_state.momenta - stepsize / 2. * init_state.grad_potential
        new_state.value = init_state.value + stepsize * p_half

        new_state.grad_potential = grad_potential(new_state.value)

        new_state.momenta = p_half - stepsize / 2. * new_state.grad_potential

        next_sample_chain = new_state.copy()
        next_sample_chain.momenta = np.vstack([p_half, new_state.momenta])
        return new_state, next_sample_chain

    final_leapfrog, all_leapfrog = scan(leapfrog_step, previous_state, None, n_steps)
    all_leapfrog.momenta = np.concatenate(all_leapfrog.momenta)

    if hasattr(previous_state, '_all_leapfrog_value'):
        all_leapfrog._all_leapfrog_value = np.concatenate([previous_state.value[np.newaxis],
                                                           all_leapfrog.value])[np.newaxis]
    if hasattr(previous_state, '_all_leapfrog_momenta'):
        all_leapfrog._all_leapfrog_momenta = np.concatenate([previous_state.momenta[np.newaxis],
                                                             all_leapfrog.momenta])[np.newaxis]

    return all_leapfrog


@decorator
def _metric_plot_decorate(plot_func: Callable,
                          *args, **kwargs) -> Union[plt.Axes, Tuple[Figure, plt.Axes]]:

    sample = args[0]
    ax = args[-1]

    if 'title' not in kwargs.keys() and hasattr(sample, 'name'):
        title = sample.name
    elif 'title' in kwargs.keys():
        title = kwargs['title']
        del kwargs['title']
    else:
        title = None

    if ax is None:
        fig, ax = plt.subplots()
        plot_func(*args[:-1], ax, **kwargs)

        if title is not None:
            ax.set_title(title)

        return fig, ax
    else:
        plot_func(*args, **kwargs)
        ax = args[-1]
        if title is not None:
            ax.set_title(title)
        return ax





from mocat import twodim

from mocat.src.scenarios.toy_scenarios import Gaussian
from mocat.src.scenarios.toy_scenarios import GaussianMixture
from mocat.src.scenarios.toy_scenarios import DoubleWell
from mocat.src.scenarios.toy_scenarios import Rastrigin






########################################################################################################################
# Module: mcmc/corrections.py
# Description: Correction mechanisms for MCMC samplers.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Tuple

from jax import random, vmap, numpy as np
from jax.lax import cond
from jax.ops import index_update

from mocat.src.core import CDict, Scenario
from mocat.src.mcmc.sampler import MCMCSampler


class Correction:

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def startup(self,
                scenario: Scenario,
                sampler: MCMCSampler):
        pass

    def correct(self,
                scenario: Scenario,
                sampler: MCMCSampler,
                reject_state: CDict,
                reject_extra: CDict,
                proposed_state: CDict,
                proposed_extra: CDict) -> Tuple[CDict, CDict]:
        raise NotImplementedError

    def __call__(self,
                 scenario: Scenario,
                 sampler: MCMCSampler,
                 reject_state: CDict,
                 reject_extra: CDict,
                 proposed_state: CDict,
                 proposed_extra: CDict) -> Tuple[CDict, CDict]:
        return self.correct(scenario, sampler,
                            reject_state, reject_extra,
                            proposed_state, proposed_extra)


class Uncorrected(Correction):

    def correct(self,
                scenario: Scenario,
                sampler: MCMCSampler,
                reject_state: CDict,
                reject_extra: CDict,
                proposed_state: CDict,
                proposed_extra: CDict) -> Tuple[CDict, CDict]:
        return proposed_state, proposed_extra


def update_ensemble_potential(scenario: Scenario,
                              state: CDict,
                              extra: CDict) -> CDict:
    ensemble_index = extra.iter % extra.parameters.n_ensemble
    state.potential = index_update(state.potential, ensemble_index,
                                   scenario.potential(state.value[ensemble_index]))
    return state


def update_potential(scenario: Scenario,
                     state: CDict,
                     extra: CDict) -> CDict:
    state.potential = scenario.potential(state.value)
    return state


class Metropolis(Correction):

    _update_potential = None

    def startup(self,
                scenario: Scenario,
                sampler: MCMCSampler):
        sampler.initial_state.alpha = 1.

        if sampler.initial_state.value.ndim == 2:
            sampler.initial_state.potential = vmap(scenario.potential)(sampler.initial_state.value)
            self._update_potential = update_ensemble_potential
        else:
            sampler.initial_state.potential = scenario.potential(sampler.initial_state.value)
            self._update_potential = update_potential

    def correct(self,
                scenario: Scenario,
                sampler: MCMCSampler,
                reject_state: CDict,
                reject_extra: CDict,
                proposed_state: CDict,
                proposed_extra: CDict) -> Tuple[CDict, CDict]:
        proposed_state = self._update_potential(scenario, proposed_state, proposed_extra)

        alpha = sampler.acceptance_probability(scenario,
                                               reject_state, reject_extra,
                                               proposed_state, proposed_extra)
        alpha = np.where(np.isnan(alpha), 0., alpha)

        random_key, subkey = random.split(proposed_extra.random_key)
        u = random.uniform(subkey)

        new_state, new_extra = cond(u < alpha,
                                    lambda _: (proposed_state, proposed_extra),
                                    lambda _: (reject_state, reject_extra),
                                    None)

        new_state.alpha = alpha
        new_extra.random_key = random_key
        return new_state, new_extra


class RMMetropolis(Metropolis):

    def __init__(self,
                 adapt_cut_off: int = np.inf,
                 rm_stepsize_scale: float = 1.,
                 rm_stepsize_neg_exponent: float = 0.75,
                 log_update: bool = True):
        super().__init__()
        self.adapt_cut_off = adapt_cut_off
        self.rm_stepsize_scale = rm_stepsize_scale
        self.rm_stepsize_neg_exponent = rm_stepsize_neg_exponent
        self.log_update = log_update
        self._param_update = self.log_robbins_monro_update if self.log_update else self.robbins_monro_update

    def startup(self,
                scenario: Scenario,
                sampler: MCMCSampler):
        # Set tuning parameter (i.e. stepsize) to 2.38^2/d if not initiated and adaptive
        if hasattr(sampler.parameters, sampler.tuning.parameter) \
                and getattr(sampler.parameters, sampler.tuning.parameter) is None:
            setattr(sampler.parameters, sampler.tuning.parameter, 2.38 ** 2 / scenario.dim)
            setattr(sampler.initial_extra.parameters, sampler.tuning.parameter, 2.38 ** 2 / scenario.dim)

        setattr(sampler.initial_state, sampler.tuning.parameter, getattr(sampler.initial_extra.parameters,
                                                                         sampler.tuning.parameter))

        super().startup(scenario, sampler)

        sampler.tuning.monotonicity = 1 if sampler.tuning.monotonicity in (1, 'increasing') else -1

        self.tuning = sampler.tuning

        self._param_update = self.log_robbins_monro_update if self.log_update else self.robbins_monro_update

    def correct(self,
                scenario: Scenario,
                sampler: MCMCSampler,
                reject_state: CDict,
                reject_extra: CDict,
                proposed_state: CDict,
                proposed_extra: CDict) -> Tuple[CDict, CDict]:
        corrected_state, corrected_extra = super().correct(scenario, sampler,
                                                           reject_state, reject_extra,
                                                           proposed_state, proposed_extra)

        adapted_state, adapted_extra = cond(corrected_extra.iter <= self.adapt_cut_off,
                                            lambda carry: self.adapt(*carry),
                                            lambda carry: carry,
                                            (corrected_state, corrected_extra))

        setattr(adapted_state, sampler.tuning.parameter, getattr(adapted_extra.parameters,
                                                                 sampler.tuning.parameter))
        return adapted_state, adapted_extra

    def adapt(self,
              state: CDict,
              extra: CDict) -> Tuple[CDict, CDict]:
        param = extra.parameters.__dict__[self.tuning.parameter]
        param = self._param_update(param, getattr(state, self.tuning.metric),
                                   self.tuning.target, self.tuning.monotonicity,
                                   self.rm_stepsize_scale
                                   * extra.iter ** -self.rm_stepsize_neg_exponent)
        extra.parameters.__dict__[self.tuning.parameter] = param
        return state, extra

    @staticmethod
    def robbins_monro_update(param, metric_value, target, monotonicity, rm_stepsize):
        return param - rm_stepsize * (metric_value - target) * monotonicity

    @staticmethod
    def log_robbins_monro_update(param, metric_value, target, monotonicity, rm_stepsize):
        log_param = np.log(param)
        new_log_param = RMMetropolis.robbins_monro_update(log_param,
                                                          metric_value,
                                                          target,
                                                          monotonicity,
                                                          rm_stepsize)
        return np.exp(new_log_param)



########################################################################################################################
# Module: mcmc/sampler.py
# Description: Generic MCMC sampler class.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Tuple, Union

from jax import numpy as np

from mocat.src.core import Sampler, Scenario, CDict


class MCMCSampler(Sampler):
    default_correction = None

    update_initial_state = True
    update_initial_extra = True

    def __init__(self,
                 **kwargs):
        self.initial_state = None
        self.initial_extra = None

        if not hasattr(self, 'parameters'):
            self.parameters = CDict()

        if not hasattr(self, 'tuning'):
            self.tuning = CDict(parameter='stepsize',
                                target=None,
                                metric='alpha',
                                monotonicity='decreasing')

        # Initiate sampler class (set any additional parameters from init)
        super().__init__(**kwargs)

    def startup(self,
                scenario: Scenario,
                random_key: np.ndarray):
        default_initial_state(self, scenario)
        default_initial_extra(self, random_key)

    def always(self,
               scenario: Scenario,
               reject_state: CDict,
               reject_extra: CDict) -> Tuple[CDict, CDict]:
        return reject_state, reject_extra

    def proposal(self,
                 scenario: Scenario,
                 reject_state: CDict,
                 reject_extra: CDict) -> Tuple[CDict, CDict]:
        raise NotImplementedError(f'{self.__class__.__name__} markov proposal not initiated')

    def proposal_potential(self,
                           scenario: Scenario,
                           reject_state: CDict, reject_extra: CDict,
                           proposed_state: CDict, proposed_extra: CDict) -> Union[float, np.ndarray]:
        raise AttributeError(f'{self.__class__.__name__} proposal_potential not initiated')

    def acceptance_probability(self,
                               scenario: Scenario,
                               reject_state: CDict, reject_extra: CDict,
                               proposed_state: CDict, proposed_extra: CDict) -> Union[float, np.ndarray]:
        raise AttributeError(f'{self.__class__.__name__} acceptance_probability not initiated')


def default_initial_state(sampler: MCMCSampler,
                          scenario: Scenario):
    if sampler.initial_state is None \
            or not hasattr(sampler.initial_state, 'value') \
            or sampler.initial_state.value.shape[-1] != scenario.dim:
        x0 = np.zeros(scenario.dim)
        sampler.initial_state = CDict(value=x0)


def default_initial_extra(sampler: MCMCSampler,
                          random_key: np.ndarray):
    if sampler.initial_extra is None:
        sampler.initial_extra = CDict(random_key=random_key,
                                      iter=0)
    if hasattr(sampler, 'parameters'):
        sampler.initial_extra.parameters = sampler.parameters


def mh_acceptance_probability(sampler: MCMCSampler,
                              scenario: Scenario,
                              reject_state: CDict, reject_extra: CDict,
                              proposed_state: CDict, proposed_extra: CDict) -> Union[float, np.ndarray]:
    pre_min_alpha = np.exp(- proposed_state.potential
                           + reject_state.potential
                           - sampler.proposal_potential(scenario,
                                                        proposed_state, proposed_extra,
                                                        reject_state, reject_extra)
                           + sampler.proposal_potential(scenario,
                                                        reject_state, reject_extra,
                                                        proposed_state, proposed_extra))

    return np.minimum(1., pre_min_alpha)



########################################################################################################################
# Module: kernels.py
# Description: Kernel class definition plus some standard kernels.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

import copy
from typing import Union
from functools import partial

import jax.numpy as np
from jax import jit

from mocat.src.core import CDict


class Kernel:

    def __init__(self, **kwargs):
        self.parameters = CDict()
        self.parameters.__dict__.update(kwargs)

    def __repr__(self):
        return f"mocat.kernels.{self.__class__.__name__}({self.__dict__.__repr__()})"

    def copy(self):
        return copy.deepcopy(self)

    def __call__(self,
                 x: np.ndarray,
                 y: np.ndarray) -> Union[float, np.ndarray]:
        raise NotImplementedError(f'kernel.{self.__class__.__name__} call not implemented')

    def grad_x(self,
               x: np.ndarray,
               y: np.ndarray) -> Union[float, np.ndarray]:
        raise NotImplementedError(f'kernel.{self.__class__.__name__} grad_x not implemented')

    def grad_y(self,
               x: np.ndarray,
               y: np.ndarray) -> Union[float, np.ndarray]:
        raise NotImplementedError(f'kernel.{self.__class__.__name__} grad_x not implemented')

    def diag_grad_xy(self,
                     x: np.ndarray,
                     y: np.ndarray) -> Union[float, np.ndarray]:
        raise NotImplementedError(f'kernel.{self.__class__.__name__} diag_grad_xy not implemented')


class Gaussian(Kernel):

    def __init__(self,
                 bandwidth: Union[float, np.ndarray] = 1.):
        super().__init__()
        self.parameters.bandwidth = bandwidth

    @partial(jit, static_argnums=(0,))
    def __call__(self,
                 x: np.ndarray,
                 y: np.ndarray) -> Union[float, np.ndarray]:
        diff = (x - y) / self.parameters.bandwidth
        return np.exp(-0.5 * np.sum(np.square(diff), axis=-1))

    @partial(jit, static_argnums=(0,))
    def grad_x(self,
               x: np.ndarray,
               y: np.ndarray) -> Union[float, np.ndarray]:
        return (y - x) * self(x, y) / self.parameters.bandwidth ** 2

    @partial(jit, static_argnums=(0,))
    def grad_y(self,
               x: np.ndarray,
               y: np.ndarray) -> Union[float, np.ndarray]:
        return (x - y) * self(x, y) / self.parameters.bandwidth ** 2

    @partial(jit, static_argnums=(0,))
    def diag_grad_xy(self,
                     x: np.ndarray,
                     y: np.ndarray) -> Union[float, np.ndarray]:
        return (self.parameters.bandwidth ** 2 - (x - y) ** 2) * self(x, y) / self.parameters.bandwidth ** 4


class IMQ(Kernel):

    def __init__(self,
                 c: float = 1.,
                 beta: float = -0.5,
                 bandwidth: Union[float, np.ndarray] = 1.):
        super().__init__()
        self.parameters.c = c
        self.parameters.beta = beta
        self.parameters.bandwidth = bandwidth

    @partial(jit, static_argnums=(0,))
    def __call__(self,
                 x: np.ndarray,
                 y: np.ndarray) -> Union[float, np.ndarray]:
        diff = (x - y) / self.parameters.bandwidth
        return np.power(self.parameters.c + 0.5 * np.sum(np.square(diff), axis=-1), self.parameters.beta)

    @partial(jit, static_argnums=(0,))
    def grad_x(self,
               x: np.ndarray,
               y: np.ndarray) -> Union[float, np.ndarray]:
        diff = (x - y) / self.parameters.bandwidth
        return diff * self.parameters.beta * np.power(self.parameters.c + 0.5 * np.sum(np.square(diff), axis=-1),
                                                      self.parameters.beta - 1)

    @partial(jit, static_argnums=(0,))
    def grad_y(self,
               x: np.ndarray,
               y: np.ndarray) -> Union[float, np.ndarray]:
        diff = (x - y) / self.parameters.bandwidth
        return - diff / self.parameters.bandwidth * self.parameters.beta\
               * np.power(self.parameters.c + 0.5 * np.sum(np.square(diff), axis=-1),
                          self.parameters.beta - 1)

    @partial(jit, static_argnums=(0,))
    def diag_grad_xy(self,
                     x: np.ndarray,
                     y: np.ndarray) -> Union[float, np.ndarray]:
        diff = (x - y) / self.parameters.bandwidth
        base = self.parameters.c + 0.5 * np.sum(np.square(diff), axis=-1)
        return (- self.parameters.beta * np.power(base, self.parameters.beta - 1)
                + self.parameters.beta * (self.parameters.beta - 1) * diff ** 2
                * np.power(base, self.parameters.beta - 2)) / self.parameters.bandwidth ** 2




########################################################################################################################
# Module: tests/test_utils.py
# Description: Tests for mocat utils
#
# Web: https://github.com/SamDuffield/mocat-jax
########################################################################################################################

import unittest

import jax.numpy as np
import numpy.testing as npt

from mocat.src import utils
from mocat.src.core import CDict


class TestLeaveOneOut(unittest.TestCase):
    n = 10

    def test_0(self):
        inds0 = utils.leave_one_out_indices(self.n, 0)
        npt.assert_array_equal(inds0, np.arange(1, self.n + 1))

    def test_2(self):
        inds2 = utils.leave_one_out_indices(self.n, 2)
        npt.assert_array_equal(inds2[:2], np.arange(2))
        npt.assert_array_equal(inds2[2:], np.arange(3, self.n + 1))

    def test_n_plus_1(self):
        inds_n_plus_1 = utils.leave_one_out_indices(self.n, self.n+1)
        npt.assert_array_equal(inds_n_plus_1, np.arange(self.n))


class TestGaussianPotential(unittest.TestCase):

    def test_n1_d1(self):
        x = np.array([7.])
        npt.assert_array_equal(utils.gaussian_potential(x), np.array(24.5))

        m = np.array([1.])
        npt.assert_array_equal(utils.gaussian_potential(x, m), np.array(18.))

        prec = np.array([[2.]])
        npt.assert_array_equal(utils.gaussian_potential(x, prec=prec), np.array(49.))
        npt.assert_array_equal(utils.gaussian_potential(x, m, prec), np.array(36.))

    def test_n1_d5(self):
        x = np.ones(5)
        npt.assert_array_equal(utils.gaussian_potential(x), np.array(2.5))

        m = 3.
        npt.assert_array_equal(utils.gaussian_potential(x, m), np.array(10.))

        m = np.ones(5) * 3.
        npt.assert_array_equal(utils.gaussian_potential(x, m), np.array(10.))

        prec = np.eye(5) * 2
        npt.assert_array_equal(utils.gaussian_potential(x, prec=prec), np.array(5.))
        npt.assert_array_equal(utils.gaussian_potential(x, m, prec), np.array(20.))

    def test_n5_d2(self):
        x = np.ones((5, 2))
        npt.assert_array_equal(utils.gaussian_potential(x), np.ones(5))

        m = 3.
        npt.assert_array_equal(utils.gaussian_potential(x, m), np.ones(5) * 4)

        m = np.ones(2) * 3.
        npt.assert_array_equal(utils.gaussian_potential(x, m), np.ones(5) * 4)

        sqrt_prec = np.array([[5., 0.], [2., 3.]])
        npt.assert_array_equal(utils.gaussian_potential(x, sqrt_prec=sqrt_prec), np.ones(5) * 29)
        npt.assert_array_equal(utils.gaussian_potential(x, m, sqrt_prec=sqrt_prec), np.ones(5) * 116)


class TestLeapfrog(unittest.TestCase):

    start_state = CDict(value=np.zeros(2),
                        grad_potential=np.array([1., 2.]),
                        momenta=np.ones(2))

    stepsize = 0.1
    n_steps = 3

    def test_leapfrog(self):
        out_state = utils.leapfrog(self.start_state,
                                   lambda x: x,
                                   self.stepsize,
                                   self.n_steps)[-1]

        npt.assert_array_equal(out_state.value, np.array([0.28120947, 0.266409]))
        npt.assert_array_equal(out_state.grad_potential, np.array([0.28120947, 0.266409]))
        npt.assert_array_equal(out_state.momenta, np.array([0.9075344, 0.8597696]))

    def test_all_leapfrog(self):
        self.start_state._all_leapfrog_value = True
        self.start_state._all_leapfrog_momenta = True

        out_state = utils.leapfrog(self.start_state,
                                   lambda x: x,
                                   self.stepsize,
                                   self.n_steps)[-1]

        self.test_leapfrog()

        npt.assert_array_equal(out_state.value, np.array([0.28120947, 0.266409]))
        npt.assert_array_equal(out_state.grad_potential, np.array([0.28120947, 0.266409]))
        npt.assert_array_equal(out_state.momenta, np.array([0.9075344, 0.8597696]))

        all_leapfrog_value = out_state._all_leapfrog_value
        all_leapfrog_momenta = out_state._all_leapfrog_momenta

        npt.assert_array_equal(all_leapfrog_value.shape, (self.n_steps + 1, 2))
        npt.assert_array_equal(all_leapfrog_momenta.shape, (2*self.n_steps + 1, 2))

        npt.assert_array_equal(all_leapfrog_value[0], np.zeros(2))
        npt.assert_array_equal(all_leapfrog_value[-1], np.array([0.28120947, 0.266409]))

        npt.assert_array_equal(all_leapfrog_momenta[0], np.ones(2))
        npt.assert_array_equal(all_leapfrog_momenta[-1], np.array([0.9075344, 0.8597696]))


if __name__ == '__main__':
    unittest.main()
